"""Rule that can be used to monitor fixation.

This rule is used to keep track of how long the subject has been fixating on a
sprite (e.g. a fixation cross). The fixation duration is stored in the
environment meta_state, and at each step either incremented if subject is
fixating or reset to 0 otherwise.

A config that uses this rule should have a conditional rule checking whether the
fixation duration in the environment meta_state has reached a threshold, in
order to proceed with the task.
"""

from . import abstract_rule
import numpy as np


class Fixation(abstract_rule.AbstractRule):
    """Fixation rule."""

    def __init__(self,
                 agent_layer,
                 fixation_layer,
                 fixation_threshold=0.1,
                 meta_state_fixation_key='fixation_duration'):
        """Constructor.

        Args:
            agent_layer: String. Key in environment state of agent layer. Agent
                is assumed to be the sprite at index 0 in this layer.
            fixation_layer: String. Key in environment state of agent layer.
                Fixation target is assumed to be the sprite at index 0 in this
                layer.
            fixation_threshold: Float. Distance from agent to fixation target
                considered fixating.
            meta_state_fixation_key: String. Key in meta_state whose value
                contains fixation duration.
        """
        self._agent_layer = agent_layer
        self._fixation_layer = fixation_layer
        self._fixation_threshold = fixation_threshold
        self._meta_state_fixation_key = meta_state_fixation_key

    def reset(self, state, meta_state):
        del state
        meta_state[self._meta_state_fixation_key] = 0

    def step(self, state, meta_state):
        agent = state[self._agent_layer][0]
        target = state[self._fixation_layer][0]
        dist = np.linalg.norm(agent.position - target.position)
        if dist < self._fixation_threshold:
            meta_state[self._meta_state_fixation_key] += 1
        else:
            meta_state[self._meta_state_fixation_key] = 0



"""Task with predators chasing agent in open arena.

The predators (red circles) chase the agent. The predators bouce off the arena
boundaries, while the agent cannot exit but does not bounce (i.e. it has
inelastic collisions with the boundaries). Trials only terminate when the agent
is caught by a predator. The subject controls the agent with a joystick.

This task also contains an auto-curriculum: When the subject does well (evades
the predators for a long time before being caught), the predators' masses are
decreased, thereby increasing the predators' speeds. Conversely, when the
subject does poorly (gets caught quickly), the predators' masses are increased,
thereby decreasing the predators' speeds.
"""

import collections
import numpy as np
import os

from moog import action_spaces
from moog import game_rules
from moog import observers
from moog import physics as physics_lib
from moog import shapes
from moog import tasks
from moog.state_initialization import distributions as distribs
from moog.state_initialization import sprite_generators


class StateInitialization():
    """State initialization class to dynamically adapt predator mass.
    
    This is essentially an auto-curriculum: When the subject does well (evades
    the predators for a long time before being caught), the predators' masses
    are decreased, thereby increasing the predators' speeds. Conversely, when
    the subject does poorly (gets caught quickly), the predators' masses are
    increased, thereby decreasing the predators' speeds.
    """

    def __init__(self, num_predators, step_scaling_factor, threshold_trial_len):
        """Constructor.

        This class uses the meta-state to keep track of the number of steps
        before the agent is caught. See the game rules section near the bottom
        of this file for the counter incrementer.

        Args:
            step_scaling_factor: Float. Fractional decrease of predator mass
                after a trial longer than threshold_trial_len. Also used as
                fractional increase of predator mass after a trial shorter than
                threshold_trial_len. Should be small and positive.
            threshold_trial_len: Length of a trial above which the predator
                mass is decreased and below which the predator mass is
                increased.
        """
        self._mass = 1.
        self._step_scaling_factor = step_scaling_factor
        self._threshold_trial_len = threshold_trial_len

        # Agent
        agent_factors = distribs.Product(
            [distribs.Continuous('x', 0., 1.),
            distribs.Continuous('y', 0., 1.)],
            shape='circle', scale=0.1, c0=0.33, c1=1., c2=0.66,
        )
        self._agent_generator = sprite_generators.generate_sprites(
            agent_factors, num_sprites=1)

        # Predators
        predator_factors = distribs.Product(
            [distribs.Continuous('x', 0., 1.),
            distribs.Continuous('y', 0., 1.)],
            shape='circle', scale=0.1, c0=0., c1=1., c2=0.8,
        )
        self._predator_generator = sprite_generators.generate_sprites(
            predator_factors, num_sprites=num_predators)

        # Walls
        self._walls = shapes.border_walls(
            visible_thickness=0., c0=0., c1=0., c2=0.5)

        self._meta_state = None

    def state_initializer(self):
        """State initializer method to be fed to environment."""
        agent = self._agent_generator(without_overlapping=self._walls)
        predators = self._predator_generator(
            without_overlapping=self._walls + agent)

        if self._meta_state is not None:
            if self._meta_state['step_count'] > self._threshold_trial_len:
                self._mass -= self._mass * self._step_scaling_factor
            else:
                self._mass += self._mass * self._step_scaling_factor
        for s in predators:
            s.mass = self._mass
        
        state = collections.OrderedDict([
            ('walls', self._walls),
            ('agent', agent),
            ('predators', predators),
        ])
        return state

    def meta_state_initializer(self):
        """Meta-state initializer method to be fed to environment."""
        self._meta_state = {'step_count': 0}
        return self._meta_state


def get_config(num_predators):
    """Get config dictionary of kwargs for environment constructor.
    
    Args:
        num_predators: Int. Number of predators.
    """

    ############################################################################
    # Sprite initialization
    ############################################################################

    state_initialization = StateInitialization(
        num_predators=num_predators,
        step_scaling_factor=0.1,
        threshold_trial_len=200,
    )

    ############################################################################
    # Physics
    ############################################################################

    agent_friction_force = physics_lib.Drag(coeff_friction=0.25)
    predator_friction_force = physics_lib.Drag(coeff_friction=0.04)
    predator_random_force = physics_lib.RandomForce(max_force_magnitude=0.03)
    predator_attraction = physics_lib.DistanceForce(
        physics_lib.linear_force_fn(zero_intercept=-0.0025, slope=0.0001))
    elastic_asymmetric_collision = physics_lib.Collision(
        elasticity=1., symmetric=False)
    inelastic_asymmetric_collision = physics_lib.Collision(
        elasticity=0., symmetric=False)
    
    forces = (
        (agent_friction_force, 'agent'),
        (predator_friction_force, 'predators'),
        (predator_random_force, 'predators'),
        (predator_attraction, 'agent', 'predators'),
        (elastic_asymmetric_collision, 'predators', 'walls'),
        (inelastic_asymmetric_collision, 'agent', 'walls'),
    )
    
    physics = physics_lib.Physics(*forces, updates_per_env_step=10)

    ############################################################################
    # Task
    ############################################################################

    task = tasks.ContactReward(
        -1, layers_0='agent', layers_1='predators', reset_steps_after_contact=0)

    ############################################################################
    # Action space
    ############################################################################

    action_space = action_spaces.Joystick(
        scaling_factor=0.01, action_layers='agent')

    ############################################################################
    # Observer
    ############################################################################

    observer = observers.PILRenderer(
        image_size=(64, 64), anti_aliasing=1, color_to_rgb='hsv_to_rgb')

    ############################################################################
    # Game rules
    ############################################################################

    def _increment_count(meta_state):
        meta_state['count'] += 1
    rules = game_rules.ModifyMetaState(_increment_count)

    ############################################################################
    # Final config
    ############################################################################

    config = {
        'state_initializer': state_initialization.state_initializer,
        'physics': physics,
        'task': task,
        'action_space': action_space,
        'observers': {'image': observer},
        'meta_state_initializer': state_initialization.meta_state_initializer,
    }
    return config



"""Rules that re-center a sprite periodically.

Typically used in first-person games to keep the agent in the middle of the
screen and move all non-agent sprites when the agent moves to preserve their
position relative to the agent.
"""

from . import abstract_rule
import numpy as np


class KeepNearCenter(abstract_rule.AbstractRule):
    """Translates many sprites at once to re-center an agent.
    
    This is used in first-person games, where there is a background grid that
    does not extend to infinity so all the other sprites must periodically be
    snapped back near the center once the agent reaches the grid size away.
    """

    def __init__(self,
                 agent_layer,
                 layers_to_center,
                 grid_x,
                 grid_y=None):
        """Constructor.

        Args:
            agent_layer: String. Must be a key in the environment state. The
            first sprite in state[agent_layer] will be kept near [0.5, 0.5].
            layers_to_center: Iterable of strings. Elements must be keys in
                environment state. All sprites in these layers will be centered
                along with the agent.
            grid_x: Maximum x-distance from 0.5 allowed by the agent before
                centering. Recentering adjusts x-positions by exactly this
                amount.
            grid_x: Maximum y-distance from 0.5 allowed by the agent before
                centering. Recentering adjusts y-positions by exactly this
                amount.
        """
        self._agent_layer = agent_layer

        if agent_layer not in set(layers_to_center):
            layers_to_center = list(layers_to_center) + [agent_layer]
        self._layers_to_center = layers_to_center
    
        if grid_y is None:
            grid_y = grid_x
        self._grid_cell = np.array([grid_x, grid_y])

    def _move_sprites(self, state, delta_pos):
        for layer in self._layers_to_center:
            for sprite in state[layer]:
                sprite.position = sprite.position + delta_pos

    def step(self, state, meta_state):
        """Run the rule, centering layers if necessary."""
        del meta_state
        
        agent = state[self._agent_layer][0]
        agent_pos = agent.position - np.array([0.5, 0.5])
        delta_pos = (
            -1. * self._grid_cell * (agent_pos > self._grid_cell) + 
            self._grid_cell * (agent_pos < -1. * self._grid_cell)
        )

        if any(delta_pos):
            self._move_sprites(state, delta_pos)



"""Gravity forces."""

from . import abstract_force
import numpy as np


class DownGravity(abstract_force.AbstractNewtonianForce):
    """Force class for downwards gravity.
    
    This gravity acts to pull objects down, as if on Earth's surface.
    """

    def __init__(self, g=-1.):
        """Constructor.

        Args:
            g: Scalar. Gravitational constant.
        """
        self._g = g

    def _compute_forces(self, sprite):
        force = self._g * sprite.mass * np.array([0, 1])
        return (force,)


class Gravity(abstract_force.AbstractNewtonianForce):
    """Force class for gravity.
    
    This gravity acts between bodies in space.
    """

    def __init__(self, g=-1., symmetric=True):
        """Constructor.

        Args:
            g: Scalar. Gravitational constant.
            symmetric: Bool. Whether to apply the force to both sprites
                involved. If False, only applies force to the second sprite, the
                second argument to .step().
        """
        self._g = g
        self._symmetric = symmetric

    def _compute_forces(self, sprite_0, sprite_1):
        """Compute forces on sprite_0 and sprite_1."""
        diff = sprite_1.position - sprite_0.position
        dist = np.linalg.norm(diff)
        if dist == 0.:
            return np.zeros(2), np.zeros(2)
        force_direction = diff / dist
        force_magnitude = self._g * sprite_0.mass * sprite_1.mass * dist

        sprite_1_force = force_magnitude * force_direction

        if self._symmetric:
            sprite_0_force = -1 * sprite_1_force
        else:
            sprite_0_force = np.array([0., 0.])

        return sprite_0_force, sprite_1_force



"""Avoid colliding predator polygons.

This task serves to showcase collisions. The predators have a variety of
polygonal shapes and bounce off each other and off the walls with Newtonian
collisions. The subject controls a green agent circle. The subject gets negative
reward if contacted by a predators and positive reward periodically.
"""

import collections
import numpy as np

from moog import action_spaces
from moog import physics as physics_lib
from moog import observers
from moog import sprite
from moog import tasks
from moog import shapes
from moog.state_initialization import distributions as distribs
from moog.state_initialization import sprite_generators


def get_config(_):
    """Get environment config."""

    ############################################################################
    # Sprite initialization
    ############################################################################

    # Agent
    agent_factors = distribs.Product(
        [distribs.Continuous('x', 0.1, 0.9),
         distribs.Continuous('y', 0.1, 0.9)],
        shape='circle', scale=0.1, c0=0.33, c1=1., c2=0.66,
    )

    # Predators
    shape_0 = 1.8 * np.array(
        [[-0.3, -0.3], [0.1, -0.7], [0.4, 0.6], [-0.1, 0.25]])
    shape_1 = 1.5 * np.array(
        [[-0.5, -0.3], [-0.1, -0.7], [0.7, 0.1], [0., -0.1], [-0.3, 0.25]])
    predator_factors = distribs.Product(
        [distribs.Continuous('x', 0.2, 0.8),
         distribs.Continuous('y', 0.2, 0.8),
         distribs.Discrete(
             'shape', [shape_0, shape_1, 'star_5', 'triangle', 'spoke_5']),
         distribs.Continuous('angle', 0., 2 * np.pi),
         distribs.Continuous('aspect_ratio', 0.75, 1.25),
         distribs.Continuous('scale', 0.1, 0.15),
         distribs.Continuous('x_vel', -0.03, 0.03),
         distribs.Continuous('y_vel', -0.03, 0.03),
         distribs.Continuous('angle_vel', -0.05, 0.05)],
        c0=0., c1=1., c2=0.8,
    )

    # Walls
    walls = shapes.border_walls(visible_thickness=0.05, c0=0., c1=0., c2=0.5)

    # Create callable initializer returning entire state
    agent_generator = sprite_generators.generate_sprites(
        agent_factors, num_sprites=1)
    predator_generator = sprite_generators.generate_sprites(
        predator_factors, num_sprites=5)

    def state_initializer():
        predators = predator_generator(
            disjoint=True, without_overlapping=walls)
        agent = agent_generator(without_overlapping=walls + predators)
        state = collections.OrderedDict([
            ('walls', walls),
            ('predators', predators),
            ('agent', agent),
        ])
        return state

    ############################################################################
    # Physics
    ############################################################################

    agent_friction_force = physics_lib.Drag(coeff_friction=0.25)
    asymmetric_collision = physics_lib.Collision(
        elasticity=1., symmetric=False, update_angle_vel=True)
    symmetric_collision = physics_lib.Collision(
        elasticity=1., symmetric=True, update_angle_vel=True)
    agent_wall_collision = physics_lib.Collision(
        elasticity=0., symmetric=False, update_angle_vel=False)
    
    forces = (
        (agent_friction_force, 'agent'),
        (symmetric_collision, 'predators', 'predators'),
        (asymmetric_collision, 'predators', 'walls'),
        (agent_wall_collision, 'agent', 'walls'),
    )
    
    physics = physics_lib.Physics(*forces, updates_per_env_step=10)

    ############################################################################
    # Task
    ############################################################################

    predator_task = tasks.ContactReward(
        -5, layers_0='agent', layers_1='predators')
    stay_alive_task = tasks.StayAlive(
        reward_period=20,
        reward_value=0.2,
    )
    task = tasks.CompositeTask(
        predator_task, stay_alive_task, timeout_steps=200)

    ############################################################################
    # Action space
    ############################################################################

    action_space = action_spaces.Joystick(
        scaling_factor=0.01, action_layers='agent')

    ############################################################################
    # Observer
    ############################################################################

    observer = observers.PILRenderer(
        image_size=(64, 64), anti_aliasing=1, color_to_rgb='hsv_to_rgb')

    ############################################################################
    # Final config
    ############################################################################

    config = {
        'state_initializer': state_initializer,
        'physics': physics,
        'task': task,
        'action_space': action_space,
        'observers': {'image': observer},
    }
    return config



"""Shapes and shape-fetching functions for common use across tasks."""

import numpy as np
from moog import polygons
from moog import sprite

# A selection of simple shapes. Elements in SHAPES can be looked up from their
# string keys in sprite.Sprite, i.e. you can give a string key as the `shape`
# argument to sprite.Sprite and it will fetch the vertices if that key is in
# this dictionary.
SHAPES = {
    'triangle': polygons.polygon(num_sides=3, theta_0=np.pi/2),
    'square': polygons.polygon(num_sides=4, theta_0=np.pi/4),
    'pentagon': polygons.polygon(num_sides=5, theta_0=np.pi/2),
    'hexagon': polygons.polygon(num_sides=6),
    'octagon': polygons.polygon(num_sides=8),
    'circle': polygons.polygon(num_sides=30),
    'star_4': polygons.star(num_sides=4, theta_0=np.pi/4),
    'star_5': polygons.star(num_sides=5, theta_0=np.pi + np.pi/10),
    'star_6': polygons.star(num_sides=6),
    'spoke_4': polygons.spokes(num_sides=4, theta_0=np.pi/4),
    'spoke_5': polygons.spokes(num_sides=5, theta_0=np.pi + np.pi/10),
    'spoke_6': polygons.spokes(num_sides=6),
}


def border_walls(visible_thickness=0.05,
                 total_thickness=0.5,
                 c0=0,
                 c1=0,
                 c2=0,
                 opacity=255):
    """Get four sprites forming a border around the [0, 1] x [0, 1] frame.

    This can be used to (i) create the visual effect of a border at the edges of
    the screen, and/or (ii) create walls around the border that can be used to
    contain sprites inside the interior of the screen.

    Args:
        visible_thickness: Float. How thick the borders within the frame should
            be.
        total_thickness: Float. How thick the border wall is in total. Depending
            on visible_thickness, much of this may lie outside of the frame. As
            long as total_thickness is greater than visible_thickness, it is not
            important. However, if visible_thickness is very small then it can
            be good to have total_thickness non-negligibly greater than zero,
            otherwise the wall sprites are extremely narrow and collisions can
            be a little unstable since their vertices and centers of mass are
            nearly collinear.
        c0: Scalar. First coordinate of color of wall sprites.
        c1: Scalar. Second coordinate of color of wall sprites.
        c2: Scalar. Third coordinate of color of wall sprites.
        opacity: Integer in [0, 255]. Opacity of wall sprites.

    Returns:
        walls: List of four sprites, the walls.
    """
    boundary_wall_shape_0 = np.array([
        [0., visible_thickness],
        [1., visible_thickness],
        [1., visible_thickness - total_thickness],
        [0., visible_thickness - total_thickness],
    ])
    distance_across_frame = 1 + total_thickness - 2 * visible_thickness
    wall_shapes = [
        boundary_wall_shape_0,
        boundary_wall_shape_0 + np.array([[0., distance_across_frame]]),
        np.flip(boundary_wall_shape_0, axis=1),
        np.flip(boundary_wall_shape_0, axis=1) + np.array(
            [[distance_across_frame, 0.]]),
    ]
    sprite_factors = dict(x=0., y=0., c0=c0, c1=c1, c2=c2, opacity=opacity)
    walls = [
        sprite.Sprite(shape=wall_shape, **sprite_factors)
        for wall_shape in wall_shapes
    ]
    return walls


def grid_lines(grid_x=0.4,
               grid_y=0.4,
               line_thickness=0.01,
               buffer_border=0.,
               c0=0,
               c1=0,
               c2=0,
               opacity=255):
    """Get grid of lines.

    Returns a list of thin rectangular sprites forming grid lines.
    
    This is sometimes used to put a grid in the background, particularly when
    using a first-person renderer for which this grid tells the player how the
    agent is moving.

    Args:
        grid_x: Float. Width of each grid cell.
        grid_y: Float. Height of each grid cell.
        line_thickness: Float. How thick the grid lines should be.
        buffer_border: Float. How far around the frame in every direction to
            create the grid. This is useful for first-person rendering, when the
            field of view sometimes extends outside [0, 1] x [0, 1].
        c0: Scalar. First coordinate of color of background grid sprites.
        c1: Scalar. Second coordinate of color of background grid sprites.
        c2: Scalar. Third coordinate of color of background grid sprites.
        opacity: Integer in [0, 255]. Opacity of background grid sprites.

    Returns:
        grid_lines: List of sprites, the grid lines.
    """
    half_num_lines_across = int(np.floor((0.5 + buffer_border) / grid_x))
    half_num_lines_up = int(np.floor((0.5 + buffer_border) / grid_y))

    x_vertices = np.linspace(
        start=0.5 - half_num_lines_across * grid_x,
        stop=0.5 + half_num_lines_across * grid_x,
        num=1 + 2 * half_num_lines_across,
    )

    y_vertices = np.linspace(
        start=0.5 - half_num_lines_up * grid_y,
        stop=0.5 + half_num_lines_up * grid_y,
        num=1 + 2 * half_num_lines_up,
    )

    sprite_factors = dict(x=0., y=0., c0=c0, c1=c1, c2=c2, opacity=opacity)
    grid_sprites = []
    def _add_sprite(min_x, max_x, min_y, max_y):
        shape = np.array([
            [min_x, min_y], [max_x, min_y], [max_x, max_y], [min_x, max_y]
        ])
        grid_sprites.append(sprite.Sprite(shape=shape, **sprite_factors))

    for x in x_vertices:
        min_x = x - 0.5 * line_thickness
        max_x = x + 0.5 * line_thickness
        min_y = -1 * buffer_border
        max_y = 1. + buffer_border
        _add_sprite(min_x, max_x, min_y, max_y)

    for y in y_vertices:
        min_x = -1 * buffer_border
        max_x = 1. + buffer_border
        min_y = y - 0.5 * line_thickness
        max_y = y + 0.5 * line_thickness
        _add_sprite(min_x, max_x, min_y, max_y)

    return grid_sprites


def circle_vertices(radius, num_sides=50):
    """Get vertices for a circle, centered about the origin.
    
    Args:
        radius: Scalar. Radius of the circle.
        num_sides: Int. Number of sides in the circle. The circle is really just
            a many-sided polygon with this many sides
        
    Returns:
        circle: Numpy array of shape [num_sides, 2] containing the vertices of
            the circle.
    """
    min_theta = 2 * np.pi / num_sides
    thetas = np.linspace(min_theta, 2 * np.pi, num_sides)
    circle = np.stack([np.sin(thetas), np.cos(thetas)], axis=1)
    circle *= radius
    return circle


def annulus_vertices(inner_radius, outer_radius, num_sides=50):
    """Get vertices for an annulus, centered about the origin.
    
    Args:
        inner_radius: Float. Radius of inner circle
        outer_radius: Float. Radius of outer circle.
        num_sides: Int. Number of sides in each circle. Each circle is really a
            many-sided polygon.

    Returns:
        annulus: Numpy array of shape [num_sides, 2] containing the vertices of
            the annulus.
    """
    inner_circle = circle_vertices(inner_radius, num_sides=num_sides)
    inner_circle = np.concatenate((inner_circle, [inner_circle[0]]), axis=0)
    outer_circle = circle_vertices(outer_radius, num_sides=num_sides)
    outer_circle = np.concatenate((outer_circle, [outer_circle[0]]), axis=0)
    annulus = np.concatenate((inner_circle, outer_circle[::-1]), axis=0)
    return annulus



"""Multi-agent cleanup task.

This task is inspired by and similar to the CleanUp task introduced in this
paper:
    "Learning Reciprocity in Complex Sequential Social Dilemmas",
    Tom Eccles, Edward Hughes, János Kramár, Steven Wheelwright, & Joel Z. Leibo
    2019, arXiv, 1903.08082

The idea of this task is that there are blue fountains at the top of the arena
and green fruit at the bottom of the arena. Each fountain can be poisoned (dull
color) or clean (bright color) and each fruit can be spoiled (dull color) or
ripe (bright color). If an agent contacts a poisoned fountain, that fountain
turns clean and a spoiled fruit turns ripe. If an agent contacts a ripe fruit,
the agent receives a reward but that fruit turns spoiled and a fountain becomes
poisoned.

So the agents have to go up to the fountains, clean them up to turn the fruit
ripe, then go down to the fruit and collect them (receiving reward), then when
there's no ripe fruit left they have to go back up to the fountains.

Some agents can be selfish free-riders by hanging out near the fruit and eating
them while others to clean the fountains. Other agents can be selfless by
cleaning fountains all the time and never eating fruit.

Note: Since this is a multi-agent task, the demo cannot be run directly on it
because the joystick only controls one agent. However, see
../../../multi_agent_example/configs/cleanup.py for a config that calls this
environment and creates hand-crafted agents so the demo will work.
"""

import collections
import numpy as np

from moog import action_spaces
from moog import game_rules
from moog import observers
from moog import physics as physics_lib
from moog import shapes
from moog import sprite
from moog import tasks
from moog.state_initialization import distributions as distribs
from moog.state_initialization import sprite_generators

_SPRITE_SCALE = 0.1

# Value color component in HSV space corresponding to clean water and ripe fruit
_GOOD_VALUE = 1.
# Value color component in HSV space corresponding to bad water and bad fruit
_BAD_VALUE = 0.3
# Anything between _GOOD_VALUE and _BAD_VALUE
_VALUE_THRESHOLD = 0.6


def get_config(_):
    """Get environment config."""

    ############################################################################
    # Sprite initialization
    ############################################################################

    # Agents
    agent_factors = distribs.Product(
        [distribs.Continuous('x', 0., 1.),
         distribs.Continuous('y', 0.35, 0.65)],
        shape='circle', scale=0.1, c1=1., c2=0.7,
    )
    agent_0_factors = distribs.Product([agent_factors], c0=0.2)
    agent_1_factors = distribs.Product([agent_factors], c0=0.1)
    agent_2_factors = distribs.Product([agent_factors], c0=0.)

    # Walls
    walls = shapes.border_walls(visible_thickness=0.05, c0=0., c1=0., c2=0.5)

    # Fountains
    fountain_factors = {
        'shape': 'circle', 'scale': 0.05, 'c0': 0.6, 'c1': 1., 'c2': _BAD_VALUE}
    fountains_across = np.linspace(0.1, 0.9, 6)
    fountains_up = np.linspace(0.75, 0.9, 2)
    fountains_grid_x, fountains_grid_y = np.meshgrid(fountains_across,
                                                     fountains_up)
    fountains_positions = zip(np.ravel(fountains_grid_x),
                              np.ravel(fountains_grid_y))
    fountain_sprites = [
        sprite.Sprite(x=x, y=y, **fountain_factors)
        for (x, y) in fountains_positions
    ]

    # Fruits
    fruit_factors = {
        'shape': 'circle', 'scale': 0.05, 'c0': 0.3, 'c1': 1., 'c2': _BAD_VALUE}
    fruits_across = np.linspace(0.1, 0.9, 6)
    fruits_up = np.linspace(0.1, 0.25, 2)
    fruits_grid_x, fruits_grid_y = np.meshgrid(fruits_across, fruits_up)
    fruits_positions = zip(np.ravel(fruits_grid_x), np.ravel(fruits_grid_y))
    fruit_sprites = [
        sprite.Sprite(x=x, y=y, **fruit_factors)
        for (x, y) in fruits_positions
    ]

    # Create callable initializer returning entire state
    agent_0_generator = sprite_generators.generate_sprites(
        agent_0_factors, num_sprites=1)
    agent_1_generator = sprite_generators.generate_sprites(
        agent_1_factors, num_sprites=1)
    agent_2_generator = sprite_generators.generate_sprites(
        agent_2_factors, num_sprites=1)

    def state_initializer():
        agent_0 = agent_0_generator(without_overlapping=walls)
        agent_1 = agent_1_generator(without_overlapping=walls)
        agent_2 = agent_2_generator(without_overlapping=walls)
        state = collections.OrderedDict([
            ('walls', walls),
            ('fountains', fountain_sprites),
            ('fruits', fruit_sprites),
            ('agent_2', agent_2),
            ('agent_1', agent_1),
            ('agent_0', agent_0),
        ])
        return state

    ############################################################################
    # Physics
    ############################################################################

    agent_friction_force = physics_lib.Drag(coeff_friction=0.25)
    asymmetric_collision = physics_lib.Collision(
        elasticity=0.25, symmetric=False)
    
    forces = (
        (agent_friction_force, ['agent_0', 'agent_1', 'agent_2']),
        (asymmetric_collision, ['agent_0', 'agent_1', 'agent_2'], 'walls'),
    )
    
    physics = physics_lib.Physics(*forces, updates_per_env_step=5)

    ############################################################################
    # Task
    ############################################################################

    task = tasks.ContactReward(
        1, layers_0='agent_0', layers_1='fruits',
        condition=lambda s_0, s_1: s_1.c2 > _VALUE_THRESHOLD,
    )

    ############################################################################
    # Action space
    ############################################################################

    action_space = action_spaces.Composite(
        agent_0=action_spaces.Joystick(
            scaling_factor=0.005, action_layers='agent_0'),
        agent_1=action_spaces.Joystick(
            scaling_factor=0.005, action_layers='agent_1'),
        agent_2=action_spaces.Joystick(
            scaling_factor=0.005, action_layers='agent_2'),
    )

    ############################################################################
    # Observer
    ############################################################################

    image_observer = observers.PILRenderer(
        image_size=(64, 64),
        anti_aliasing=1,
        color_to_rgb='hsv_to_rgb',
    )
    raw_state_observer = observers.RawState()  # needed by hand-crafted agents

    ############################################################################
    # Game rules
    ############################################################################

    def _spoil_fruit(sprite):
        sprite.c2 = _BAD_VALUE
    def _ripen_fruit(sprite):
        sprite.c2 = _GOOD_VALUE
    def _poison_fountain(sprite):
        sprite.c2 = _BAD_VALUE
    def _clean_fountain(sprite):
        sprite.c2 = _GOOD_VALUE

    def agents_contacting_layer(state, layer, value):
        n_contact = 0
        for s in state[layer]:
            if s.c2 != value:
                continue
            n_contact += (
                s.overlaps_sprite(state['agent_0'][0]) or 
                s.overlaps_sprite(state['agent_1'][0]) or 
                s.overlaps_sprite(state['agent_2'][0])
            )
        return n_contact
    
    poison_fountains = game_rules.ModifySprites(
        layers='fountains', modifier=_poison_fountain, sample_one=True,
        filter_fn=lambda s: s.c2 > _VALUE_THRESHOLD)
    poison_fountains = game_rules.ConditionalRule(
        condition=lambda s: agents_contacting_layer(s, 'fruits', _GOOD_VALUE),
        rules=poison_fountains,
    )
    ripen_fruits = game_rules.ModifySprites(
        layers='fruits', modifier=_ripen_fruit, sample_one=True,
        filter_fn=lambda s: s.c2 < _VALUE_THRESHOLD)
    ripen_fruits = game_rules.ConditionalRule(
        condition=lambda s: agents_contacting_layer(s, 'fountains', _BAD_VALUE),
        rules=ripen_fruits,
    )

    spoil_fruits = game_rules.ModifyOnContact(
        layers_0='fruits',
        layers_1=('agent_0', 'agent_1', 'agent_2'),
        modifier_0=_spoil_fruit,
        filter_0=lambda s: s.c2 > _VALUE_THRESHOLD)
    clean_fountains = game_rules.ModifyOnContact(
        layers_0='fountains',
        layers_1=('agent_0', 'agent_1', 'agent_2'),
        modifier_0=_clean_fountain,
        filter_0=lambda s: s.c2 < _VALUE_THRESHOLD)
    
    rules = (poison_fountains, spoil_fruits, ripen_fruits, clean_fountains)

    ############################################################################
    # Final config
    ############################################################################

    config = {
        'state_initializer': state_initializer,
        'physics': physics,
        'task': task,
        'action_space': action_space,
        'observers': {'image': image_observer, 'state': raw_state_observer},
        'game_rules': rules,
    }
    return config



"""Two-body distance-based forces."""

from . import abstract_force
import abc
import numpy as np


class DistanceForce(abstract_force.AbstractNewtonianForce,
                    metaclass=abc.ABCMeta):
    """Distance force base class.
    
    This class can be used for any force that attracts or repels two sprites as
    a function of their distance alone.
    """

    def __init__(self, force_fn, symmetric=False):
        """Constructor

        Args:
            force_fn: Force function that takes a scalar distance and produces a
                scalar force magnitude, the magnitude of the force to be applied
                between sprites at the given distance apart.
            symmetric: Bool. Whether to apply the force to both sprites
                involved. If False, only applies force to the second sprite, the
                second argument to .step().
        """
        self._force_fn = force_fn
        self._symmetric = symmetric

    def _compute_forces(self, sprite_0, sprite_1):
        """Compute forces on sprite_0 and sprite_1."""
        diff = sprite_1.position - sprite_0.position
        dist = np.linalg.norm(diff)
        if dist == 0.:
            return np.zeros(2), np.zeros(2)
        force_direction = diff / dist
        force_magnitude = self._force_fn(dist)
        sprite_1_force = force_magnitude * force_direction

        if self._symmetric:
            sprite_0_force = -1 * sprite_1_force
        else:
            sprite_0_force = np.array([0., 0.])

        return sprite_0_force, sprite_1_force


def linear_force_fn(zero_intercept,
                    slope,
                    apply_distant_force=False,
                    apply_nearby_force=True):
    """Force is a linear function of distance.
    
    Args:
        zero_intercept: Scalar. Magnitude of force between sprites a distance
            zero apart.
        slope: Scalar. Slope of linear change in force with respect to distance.
        apply_distant_force: Bool. Whether to apply force at distances greater
            than the event horizon (distance at which the force is zero).
        apply_distant_force: Bool. Whether to apply force at distances less than
            the event horizon.

    Returns:
        force_fn: Function distance -> force magnitude.
    """
    event_horizon = -1. * zero_intercept / slope
    def force_fn(distance):
        force_magnitude = zero_intercept + slope * distance
        if not apply_distant_force and distance > event_horizon:
            force_magnitude = 0
        if not apply_nearby_force and distance < event_horizon:
            force_magnitude = 0
        return force_magnitude
    return force_fn


def spring_force_fn(spring_constant, equilibrium=0):
    """Spring force by Hooke's Law.
    
    Args:
        spring_constant: Scalar. Spring constant.
        equilibrium: Spring equilibrium. Distance at which the force is zero.

    Returns:
        force_fn: Function distance -> force magnitude.
    """
    def force_fn(distance):
        return -1. * spring_constant * (distance - equilibrium)
    return force_fn



"""Composite action space that composes multiple action spaces.

This is often used for multi-agent games and for games where the subject gives
multiple sources of control, e.g. eye position and joystick movement.
"""

from . import abstract_action_space


class Composite(abstract_action_space.AbstractActionSpace):
    """Composite action space.
    
    Example usage:
        joystick_action_space = maze_action_space.Joystick(
            scaling_factor=0.1, action_layers='agent')
        eye_action_space = action_spaces.SetPosition(
            action_layers='eye_sprite')
        
        action_space = action_spaces.Composite(
            joystick=joystick_action_space,
            eye=eye_action_space,
        )

        # Playing the game...
        action = {'joystick': [0.1, -0.2], 'eye': [0.4, 0.8]}
        env.step(action=action)  # goes to action_space.step(state, action)
    """

    def __init__(self, **action_spaces):
        """Constructor.
        
        Args:
            action_spaces: Dict. Keys are strings and values are action spaces.
                An action is a dictionary with the same set of keys.
        """
        self.action_spaces = action_spaces
        self._action_keys = action_spaces.keys()

        self._action_spec = {
            key: value.action_spec() for key,value in self.action_spaces.items()
        }

    def step(self, state, action):
        """Apply action to environment state.

        Args:
            state: Ordereddict of layers of sprites. Environment state.
            action: Dict. Keys much be the same as self._action_keys. Each value
                will be fed into the action space of the corresponding key.
        """
        for k, v in action.items():
            self.action_spaces[k].step(state, v)

    def reset(self, state):
        for k in self.action_spaces:
            self.action_spaces[k].reset(state)

    def random_action(self):
        """Return randomly sampled action."""
        random_action = {
            k: self.action_spaces[k].random_action() for k in self._action_keys}
        return random_action

    def action_spec(self):
        return self._action_spec

    @property
    def action_keys(self):
        return list(self._action_keys)



"""Friction forces.

This file has two classes, KineticFriction and Drag. The only difference between
the two is that kinetic friction is velocity-independent, whereas drag scales
linearly with velocity. For joystick control of agents, using Drag is a bit
easier since it prevents the velocity from exploding.
"""

from . import abstract_force
import numpy as np


class KineticFriction(abstract_force.AbstractNewtonianForce):
    """Kinetic friction force class."""

    def __init__(self, coeff_friction=1.):
        """Constructor.

        Args:
            coeff_friction: Scalar. Coefficient of friction. The frictional
                force applied to a sprite is -coeff_friction * sprite.mass.
        """
        self._coeff_friction = coeff_friction

    def _compute_forces(self, sprite):
        velocity_norm = np.linalg.norm(sprite.velocity)
        if velocity_norm == 0:
            normalized_velocity = np.zeros(2)
        else:
            normalized_velocity = sprite.velocity / velocity_norm
        force = -1 * self._coeff_friction * normalized_velocity * sprite.mass

        return (force,)
        

class Drag(abstract_force.AbstractNewtonianForce):
    """Drag force class.
    
    Note: The drag force implemented here depends only on {velocity, mass,
    coef_friction}, not on cross sectional area, so it is not a realistic drag
    force.
    """

    def __init__(self, coeff_friction=1.):
        """Constructor.

        Args:
            coeff_friction: Scalar. Coefficient of friction. The frictional
                force applied to a sprite is
                -coeff_friction * sprite.velocity * sprite.mass.
        """
        self._coeff_friction = coeff_friction

    def _compute_forces(self, sprite):
        force = -1 * self._coeff_friction * sprite.velocity * sprite.mass
        return (force,)



"""Rules that make sprites vanish."""

from . import abstract_rule
from .contact_rules import get_contact_indices
import abc
import numpy as np


class Vanish(abstract_rule.AbstractRule, metaclass=abc.ABCMeta):
    """Metaclass for rules that make sprites vanish."""

    def __init__(self, layer):
        """Constructor.

        Args:
            layer: String. Must be a key in the environment state. Sprites in
                this layer will be removed from the state if they are indexes by
                self._get_vanish_inds().
        """
        self._layer = layer

    @abc.abstractmethod
    def _get_vanish_inds(self, state):
        """Takes in state and returns an iterable of indices.
        
        Returned indices must index which elements of state[self._layer] should
            vanish.
        """
        pass

    def step(self, state, meta_state):
        """Remove sprites in specified indices of vanishing layer."""
        del meta_state
        
        vanish_inds = self._get_vanish_inds(state)
        count_vanished_already = 0
        for i in vanish_inds:
            state[self._layer].pop(i - count_vanished_already)
            count_vanished_already += 1


class VanishByFilter(Vanish):
    """Makes a sprite vanish based on a boolean function of itself."""

    def __init__(self, layer, filter_fn=None):
        """Constructor.

        Args:
            layer: String. Must be a key in the environment state. Sprites in
                this layer will be removed from the state if filter_fn returns
                True on them.
            filter_fn: None or Function taking in a sprite and return a bool
                indicating whether the given sprite should vanish. If None,
                defaults to True, i.e. vanishing all sprites in layer.
        """
        super(VanishByFilter, self).__init__(layer)
        if filter_fn is None:
            self._filter_fn = lambda _: True
        else:
            self._filter_fn = filter_fn

    def _get_vanish_inds(self, state):
        vanish_inds = np.argwhere(
            [self._filter_fn(s) for s in state[self._layer]])[:, 0]
        return vanish_inds


class VanishOnContact(Vanish):
    """Makes a sprite vanish if it is in contact with another sprite."""

    def __init__(self, vanishing_layer, contacting_layer):
        """Constructor.

        Args:
            vanishing_layer: String. Must be a key in the environment state.
                Sprites in this layer will be removed from the state if they
                contact a sprite in contacting_layer.
            contacting_layer: String. Must be a key in the environment state.
        """
        super(VanishOnContact, self).__init__(vanishing_layer)
        self._get_contact_indices = get_contact_indices(
            vanishing_layer, contacting_layer)

    def _get_vanish_inds(self, state):
        contact_inds = self._get_contact_indices(state)
        return np.unique([i for (i, j) in contact_inds])



"""Abstract task."""

import abc


class AbstractTask(abc.ABC):
    """Abstract task class.
    
    All tasks must inherit from this class.
    """

    @abc.abstractmethod
    def reset(self, state, meta_state):
        """Reset the task.

        This should do whatever the task needs to do to itself when the
        environment resets. For example, this could reset a count of the number
        of steps taken in the episode, zero out any running state information,
        etc.
        
        Args:
            state: OrderedDict of iterables of sprites.
        """
        pass
    
    @abc.abstractmethod
    def reward(self, state, meta_state, step_count):
        """Get reward and whether to reset based on the environment state.
        
        Args:
            state: OrderedDict of iterables of sprites.
            step_count: Current number of environment steps since last reset.

        Returns:
            reward: Scalar. Reward for current state.
            should_reset: Bool. Whether environment should reset on next step.
        """
        pass



"""Maze physics.

The MazePhysics class in this file is a physics object that forces sprites in
specified layers to move on a grid in a maze.

It is typically used as corrective physics for a physics.Physics instance.
"""

import numpy as np
from moog import maze_lib
from moog import physics as physics_lib

# Small imprecision tolerance when determining whether a sprite is on a grid
# line in a maze.
_EPSILON = 1e-5


class MazePhysics(physics_lib.AbstractPhysics):
    """Maze physics class."""

    def __init__(self, maze_layer='walls', avatar_layers=(),
                 constant_speed=None, max_speed=None):
        """Constructor.

        This class constrains sprites in avatar_layers to move in the maze
        specified by the sprites in maze_layer.
        
        Args:
            maze_layer: String. Name of the layer in the environment containing
                the maze wall sprites.
            avatar_layers: Iterable of strings. Sprites in these layers are
                constrained to move in the maze.
            constant_speed. None or float. If Float, all sprites in
                avatar_layers move with this constant speed.
            max_speed: None or float. If float, all sprites in avatar_layers
                move no faster than max_speed. If None, then no speed limit. If
                constant_speed is not None, then max_speed is ignored.
        """
        super(MazePhysics, self).__init__(updates_per_env_step=1)
        self._maze_layer = maze_layer
        self._avatar_layers = avatar_layers
        self._constant_speed = constant_speed
        self._max_speed = max_speed

    def reset(self, state):
        """Resetting re-infers the maze from the state."""
        self._maze = maze_lib.Maze.from_state(
            state, maze_layer=self._maze_layer)

    def _get_position_affordances(self, position):
        """Get affordances of a position.
        
        The affordances of a position indicate how far one can travel from that
        position in each direction.

        Args:
            position: Numpy float array of size (2,).
        
        Returns:
            position: Numpy float array of size (2,). The input position,
                possibly perturbed if any coordinates are within _EPSILON of
                grid lines. This rounding (essentially snapping the position to
                the grid if it's close) ensures that numerical instabilities
                don't accumulate.
            affordances: Numpy float array of size (2, 2). affordances[:, 0]
                indicate how far the position could be moved in the negative
                direction for each axis before hitting an intersection or going
                off the maze. affordances[:, 1] are similiar for the positive
                direction.
        """
        grid_side = self._maze.grid_side
        half_grid_side = self._maze.half_grid_side

        # Figure out which axes are on grid lines, and snap position to grid
        nearest_inds = (np.round(position / grid_side - 0.5)).astype(int)
        rounded_position = half_grid_side + nearest_inds * grid_side
        on_grid = np.abs(rounded_position - position) < _EPSILON
        new_position = np.copy(position)
        new_position[on_grid] = rounded_position[on_grid]

        # Get the affordances
        affordances = np.zeros((2, 2))
        inds = ((position - half_grid_side) // grid_side).astype(int)
        inds[on_grid] = nearest_inds[on_grid]
        if not any(on_grid):  # This should never happen
            raise ValueError(
                'Object is not on the maze grid. This could happen if you '
                'initialized or somehow forced a sprite position to lie off '
                'the maze grid. This could also happen if you are using a '
                'corrective_physics after the MazePhysics, in which case that '
                'later corrective_physics is adjusting the velocities produced '
                'by the MazePhysics and making the sprites run off the maze. '
                'This is bad --- if MazePhysics is used, it must be the last '
                'corrective_physics object in the physics, so that the '
                'velocities it produces are immediately enacted.')
        elif all(on_grid):  # The position lies at a grid vertex
            valid_directions = self._maze.valid_directions(inds[0], inds[1])
            affordances = valid_directions * grid_side * np.array(
                [[-1., 1.], [-1., 1.]])
        else:  # The position lies on a grid edge
            i = 1 - np.argwhere(on_grid)[0][0]
            lower = inds[i] * grid_side + half_grid_side - position[i]
            upper = (inds[i] + 1) * grid_side + half_grid_side - position[i]
            affordances[i] = np.array([lower, upper])

        return position, affordances            

    def _get_new_velocity(self, position, velocity, affordances, axis=None):
        """Get the maze-adjusted velocity.

        Args:
            position: Numpy float array of size (2,).
            velocity: Numpy float array of size (2,).
            affordances: Numpy float array of size (2, 2) containing how far one
                can travel from position in each direction.
            axis: Axis on which to travel. If None, uses the highest-speed axis.
        """
        if axis is None:  # Find the highest-speed axis
            axis = np.argmax(np.abs(velocity))
            
        if affordances[axis, 0] <= velocity[axis] <= affordances[axis, 1]:
            # Can travel with velocity along axis without hitting walls or
            # intersections
            velocity[1 - axis] = 0
            return velocity
        else:
            direction = int(0.5 + 0.5 * np.sign(velocity[axis]))
            if affordances[axis, direction] == 0:
                # We cannot move in the axis direction, so resort to other axis
                axis = 1 - axis
                direction = int(0.5 + 0.5 * np.sign(velocity[axis]))
                if affordances[axis, direction] == 0 or velocity[axis] == 0:
                    # Cannot move anywhere
                    return np.zeros(2)
                else:  # Move along other axis
                    return self._get_new_velocity(
                        position, velocity, affordances, axis=axis)
            else:  # Affordances let us reach a vertex
                # Compute the affordances at the vertex
                position[axis] += affordances[axis, direction]
                _, vertex_affordances = self._get_position_affordances(position)

                # Compute the remaining velocity at the vertex
                scaling = affordances[axis, direction] / velocity[axis]
                velocity_remainder = (1. - scaling) * np.copy(velocity)

                # Compute the velocity after hitting the vertex
                velocity_post_vertex = self._get_new_velocity(
                    position, velocity_remainder, vertex_affordances)
                
                # Update velocity to travel to the vertex then add the
                # post-vertex velocity
                velocity *= scaling
                velocity[1 - axis] = 0
                velocity += velocity_post_vertex

                return velocity

    def _update_sprite_angle(self, sprite, new_velocity):
        """Update sprite angle if necessary.
        
        This makes sprites rotate as they take turns in a maze.
        """
        # We have a little case-work to do here, because of the possibility of
        # some velocity components being zero
        if new_velocity[0] == 0 and new_velocity[1] == 0:
            new_angle = np.nan
        elif new_velocity[1] == 0:
            new_angle = -0.5 * np.sign(new_velocity[0]) * np.pi
        else:
            if np.sign(new_velocity[1]) > 0:
                new_angle = np.arctan(-new_velocity[0] / new_velocity[1])
            else:
                new_angle = np.pi + np.arctan(
                    -new_velocity[0] / new_velocity[1])
        
        # Update the sprite's angle
        if (not np.isnan(new_angle) and
                np.abs(new_angle - sprite.angle) > _EPSILON):
            sprite.angle = new_angle
                
    def _update_sprite_in_maze(self, sprite):
        """Update a sprite velocity to abide by the maze."""
        position = sprite.position
        velocity = sprite.velocity
        
        if np.all(velocity == 0) or np.any(np.isnan(velocity)):
            return

        if self._max_speed is not None:
            velocity = np.clip(velocity, -self._max_speed, self._max_speed)
        
        if self._constant_speed is not None:
            velocity += np.sign(velocity)
            velocity *= self._constant_speed

        position, affordances = self._get_position_affordances(position)
        sprite.position = np.copy(position)
        new_velocity = self._get_new_velocity(position, velocity, affordances)
        
        # Update sprite angle and velocity
        self._update_sprite_angle(sprite, new_velocity)
        sprite.velocity = new_velocity

    def apply_physics(self, state, updates_per_env_step):
        """Move the sprites according to the physics."""
        if updates_per_env_step != 1:
            raise ValueError('Must have updates_per_env_step be 1 for maze.')
        for layer in self._avatar_layers:
            for sprite in state[layer]:
                self._update_sprite_in_maze(sprite)



"""Human gui to demo tasks.

This contains tools for an interactive gui controlled by a human. The main class
is HumanAgent, which takes and environment and runs it in an iteractive display
that shows the rendered environement, plot of recent rewards, and a
human-playable action space (e.g. cartoon joystick, arrow keys, etc.).

This is useful for testing task prototypes.

Note: When recording gifs, be sure to pressing the 'esc' key when you want to
stop the demo and write the gif.

Note: If the reward plot does not appear in your figure window, that is probably
because your monitor screen is not tall enough to fit it, given the rendering
size you chose. Consider using a smaller render_size in the renderer (eg. 256).
"""

import logging
import math
import numpy as np
import sys
import time
import tkinter as tk

import mss
from matplotlib import pyplot as plt
from PIL import Image
from PIL import ImageTk
from matplotlib.backends import backend_tkagg

from moog_demos import gui_frames
from moog import action_spaces

_WINDOW_ASPECT_RATIO = 2.7  # height/width for the gui window


class HumanAgent():
    """Human-playable agent.

    This provides a gui for human interaction with an environment. The gui
    displays the rendered environment, a plot of rewards over recent history,
    and a frame which may contain a joystick, depending on the action space.

    Note: This agent does not abide by the standard RL agent api in that it does
    not have a .step() method taking an observation and returning an action.
    This is because due to Tkinter's interactive .mainloop() function, the
    entire environment interaction must be implemented as a callback in the gui.
    """

    def __init__(self, env, render_size, fps=10, reward_history=20,
                 observation_image_key='image', gif_writer=None,
                 reward_border_width=10):
        """Constructor.

        Args:
            env: instance of moog.environment.Environment. The environment
                observations are assumed to have a key 'image' whose value is an
                image with height and width render_size.
            render_size: Int. Width and height of observation_image_key value of
                the environment observation image.
            fps: Int. Frames per second to run, if possible. Note: The tkinter
                gui interface and matplotlib rendering is slow, often taking
                about 40 milliseconds for 256x256 rendering. Furthermore,
                depending on the environment and physics, stepping the
                environent can take some time (usually no more than 10
                milliseconds). Therefore, the fastest fps this gui can run is
                20fps for 256x256 rendering, and if the given fps is higher than
                this it will be capped. This slowness is mostly due to
                matplotlib and tkinter, not the environment or rendering itself,
                and does not occur when training agents or using mworks for
                psychophysics.
            reward_history: Int. Number of history timesteps to plot the reward
                for.
            observation_image_key: String. Key of the observation that is the
                image.
            gif_writer: Optional instance of a gif writer. This writer should
                have a .close() method and an .add(frame) method.
            reward_border_width: Int. Width of the red/green border to render
                when a positive/negative reward is given.
        """
        self._env = env
        self._ms_per_step = 1000. / fps
        self._reward_history = reward_history
        self._observation_image_key = observation_image_key
        self._gif_writer = gif_writer
        self._reward_border_width = reward_border_width

        # This will be used later to capture a section of the screen
        if self._gif_writer:
            self._screen_capture = mss.mss()

        # Create root Tk window and fix its size
        self.root = tk.Tk()
        frame_width = str(render_size)
        frame_height = str(int(_WINDOW_ASPECT_RATIO * render_size))
        self.root.geometry(frame_width + 'x' + frame_height)

        # Bind escape key to terminate gif_writer and exit
        def _escape(event):
            if self._gif_writer:
                self._gif_writer.close()
            sys.exit()

        self.root.bind('<Escape>', _escape)

        ########################################################################
        # Create the environment display and pack it into the top of the window.
        ########################################################################

        image = env.reset().observation[self._observation_image_key]
        self._env_canvas = tk.Canvas(
            self.root, width=render_size, height=render_size)
        self._env_canvas.pack(side=tk.TOP)
        img = ImageTk.PhotoImage(image=Image.fromarray(image))
        self._env_canvas.img = img
        self.image_on_canvas = self._env_canvas.create_image(
            0, 0, anchor="nw", image=self._env_canvas.img)

        ########################################################################
        # Create the gui frame and pack it into the bottom of the window.
        ########################################################################

        canvas_half_width = render_size / 2
        action_space = env.action_space
        if isinstance(action_space, action_spaces.Composite):
            num_agents = len(action_space.action_spaces)
            a_spaces = action_space.action_spaces.values()
            if (num_agents == 2 and all(
                    isinstance(a, action_spaces.Grid) for a in a_spaces)):
                logging.info(
                    '2-player Grid action space. One player uses arrow keys '
                    'and the other uses [a, s, d, w].')
                # Two-player Grid action spaces
                self.gui_frame = gui_frames.TwoPlayerGridActions(
                    self.root,
                    canvas_half_width=canvas_half_width,
                    player_0=action_space.action_keys[0],
                    player_1=action_space.action_keys[1],
                )
            elif (num_agents == 2 and all(
                    isinstance(a, action_spaces.Joystick) for a in a_spaces)):
                logging.info(
                    '2-player Joystick action space. One player uses arrow keys '
                    'and the other uses [a, s, d, w].')
                # Two-player Joystick action spaces
                self.gui_frame = gui_frames.TwoPlayerJoystick(
                    self.root,
                    canvas_half_width=canvas_half_width,
                    player_0=action_space.action_keys[0],
                    player_1=action_space.action_keys[1],
                )
            else:
                logging.info(
                    'Composite action space provided, human controls only the '
                    'first agent.')
                action_space = list(action_space.action_spaces.values())[0]

        if isinstance(action_space, action_spaces.Joystick):
            logging.info(
                'Joystick action space, drag and move the joystick at the '
                'bottom of the window.')
            self.gui_frame = gui_frames.JoystickFrame(
                self.root,
                canvas_half_width=canvas_half_width,
                motion_zone_radius=canvas_half_width - 5,
            )
        elif isinstance(action_space, action_spaces.Grid):
            logging.info('Grid action space, use arrow keys.')
            self.gui_frame = gui_frames.GridActions(
                self.root,
                canvas_half_width=canvas_half_width,
            )
        elif isinstance(action_space, action_spaces.SetPosition):
            logging.info('SetPosition action space, click on the frame to act.')
            self.gui_frame = gui_frames.SetPositionFrame(
                self._env_canvas,
                canvas_half_width=canvas_half_width,
            )
        elif not isinstance(action_space, action_spaces.Composite):
            raise ValueError(
                'Cannot demo action space {}'.format(env.action_space))

        if not isinstance(action_space, action_spaces.SetPosition):
            self.gui_frame.canvas.pack(side=tk.BOTTOM)

        ########################################################################
        # Create the reward plot and pack it into the middle of the window.
        ########################################################################

        # This figuresize and the subplot adjustment is hand-crafted to make it
        # look okay for _WINDOW_ASPECT_RATIO 2.7. Ideally, we would infer the
        # space available for this reward plot and size the figure accordingly,
        # but I could not easily figure out how to do that --- it seems like
        # matplotlib doesn't count axis ticks and labels as part of the figure
        # size so those are cut off without handcrafting some subplot
        # adjustment.
        fig = plt.Figure(figsize=(2, 2), dpi=100)
        fig.subplots_adjust(bottom=0.5, left=0.25, right=0.95)

        self._ax_reward = fig.add_subplot(111)
        self._ax_reward.set_ylabel('Reward')
        self._ax_reward.set_xlabel('Time')
        self._ax_reward.axhline(y=0.0, color='lightgrey')

        # Plot rewards as a bar plot
        self._reset_rewards()
        reward_ticks = np.arange(-1 * self._reward_history + 1, 1)
        self.rewards_plot = self._ax_reward.bar(reward_ticks, self._rewards)

        # Create canvas in which to draw the reward plot and pack it
        # A tk.DrawingArea.
        self.rewards_canvas = backend_tkagg.FigureCanvasTkAgg(
            fig, master=self.root)
        self.rewards_canvas.draw()
        self.rewards_canvas.get_tk_widget().pack(side=tk.TOP, fill=tk.BOTH)

        ########################################################################
        # Start run loop, automatically running the environment.
        ########################################################################

        self.root.after(math.floor(self._ms_per_step), self.step)
        self.root.mainloop()

    def _reset_rewards(self):
        self._rewards = np.zeros(self._reward_history)
        self._reward_range = [-1e-3, 1e-3]

    def render(self, observation):
        """Render the environment display and reward plot."""
        # Put green border if positive reward, red border if negative reward
        observation_image = observation.observation[self._observation_image_key]

        if self._reward_border_width:
            # Add a red/green border to the image for positive/negative reward.
            if sum(self._rewards[-4:]) > 0:
                show_border = True
                border_color = np.array([0, 255, 0], dtype=np.uint8)
            elif sum(self._rewards[-4:]) < 0:
                show_border = True
                border_color = np.array([255, 0, 0], dtype=np.uint8)
            else:
                show_border = False

            if show_border:
                observation_image[:self._reward_border_width] = border_color
                observation_image[-self._reward_border_width:] = border_color
                observation_image[:, :self._reward_border_width] = border_color
                observation_image[:, -self._reward_border_width:] = border_color

        # Set the image in the environment display to the new observation
        self._env_canvas.img = ImageTk.PhotoImage(Image.fromarray(
            observation_image))
        self._env_canvas.itemconfig(
            self.image_on_canvas, image=self._env_canvas.img)

        # Set the reward plot data to the current self._rewards
        for rect, h in zip(self.rewards_plot, self._rewards):
            rect.set_height(h)
            rect.set_facecolor('g' if h > 0 else 'r')
        self.rewards_canvas.draw()
        self._ax_reward.set_ylim(*self._reward_range)

    def step(self):
        """Take an action in the environment and render."""
        step_start_time = time.time()
        action = self.gui_frame.action  # action from the gui
        observation = self._env.step(action)

        # Update rewards and reward_range
        self._rewards[:-1] = self._rewards[1:]
        self._rewards[-1] = observation.reward
        if observation.reward:
            self._reward_range[0] = min(
                self._reward_range[0], observation.reward)
            self._reward_range[1] = max(
                self._reward_range[1], observation.reward)

        # display new observation
        self.render(observation)
        if observation.last():
            self._reset_rewards()

        # Screengrab the window and update the gif_writer
        if self._gif_writer:
            window = {
                'top': self.root.winfo_rooty(),
                'left': self.root.winfo_rootx(),
                'width': int(self.root.winfo_width()),
                'height': int(self.root.winfo_height()),
            }
            img = np.asarray(self._screen_capture.grab(window))
            # For some reason screen capture switches red and blue color
            # channels
            img = img[:, :, [2, 1, 0, 3]]
            self._gif_writer.add(img)

        # Recurse to step again after self._ms_per_step milliseconds
        step_end_time = time.time()
        delay = (step_end_time - step_start_time) * 1000  # convert to ms
        self.root.after(math.floor(self._ms_per_step - delay), self.step)



"""This file contains functions to randomly generate mazes.

The main function in this file is generate_random_maze_matrix(), which generates
a maze with no dead ends and no open squares.
"""

import numpy as np

# Maximum iteration through a while loop
_MAX_ITERS = int(1e5)


def _get_neighbors(size, point):
    """Get indices of point's neighbors in square matrix of size `size`.

    Unless point (i, j) is on the boundary of the size x size square, this will
    be a list of 4 elements.

    Args:
        size: Int.
        point: Tuple of ints (i, j). Must satisfy 0 <= i, j < size.

    Returns:
        neighbors: List of tuples. Length 2 (if point is a corner), 3 (if point
            is on an edge), or 4 (if point is in the interior).
    """
    i, j = point

    neighbors = [(i - 1, j), (i + 1, j), (i, j - 1), (i, j + 1)]
    _valid_neighbor = lambda neighbor: all(0 <= x < size for x in neighbor)
    neighbors = list(filter(_valid_neighbor, neighbors))
    
    return neighbors


def _get_containing_blocks(size, point):
    """Get 2x2 blocks containing point in open maze of size `size`.

    Unless point is on the boundary of the size x size square, there will be 4
    containing 2x2 blocks.

    Args:
        size: Int.
        point: Tuple of ints (i, j). Must satisfy 0 <= i, j < size.

    Returns:
        block_inds: List of tuples. If (k, l) is in block_inds, then point
            (i, j) is in {(k, l), (k + 1, l), (k, l + 1), (k + 1, l + 1)}.
    """
    i, j = point
    block_inds = []
    if i > 0:
        if j > 0:
            block_inds.append((i - 1, j - 1))
        if j < size - 1:
            block_inds.append((i - 1, j))
    if i < size - 1:
        if j > 0:
            block_inds.append((i, j - 1))
        if j < size - 1:
            block_inds.append((i, j))
            
    return block_inds


def _remove_dead_ends(maze):
    """Iteratively remove the dead ends in the maze.

    This is done in place, and by the time this function returs there will be no
    open points with fewer than 2 open neighbors, i.e. no dead ends in the maze.

    Args:
        maze: N x N binary matrix.
    """
    def _fill_maze():
        # Fill in dead ends, return True if the maze has no dead ends, otherwise
        # False.
        size = maze.shape[0]
        for i in range(size):
            for j in range(size):
                if maze[i, j]:  # Not an open point
                    continue
                num_open_neighbors = np.sum(
                    [1 - maze[n[0], n[1]]
                    for n in _get_neighbors(size, (i, j))])
                if num_open_neighbors < 2:
                    maze[i, j] = 1
                    return False
        return True

    valid_maze = False
    while not valid_maze:
        valid_maze = _fill_maze()
    

def generate_random_maze_matrix(size, ambient_size=None):
    """Generate a random maze matrix.

    The maze matrix generated has no open points (e.g. no four open cells
    sharing a vertex), no dead ends (e.g. each open point has at least two open
    neighbors), and is one connected component.

    The way this generator works is it starts will a single open cell (all other
    cells are walls). Then it iteratively adds closed neighbors, as long as
    opening them doesn't open up a block. Once there are no more such neighbors,
    it iteratively fills in all dead ends. The result is the final maze matrix
    (unless it is all walls, in which case the function recurses to try again).

    Args:
        size: Int. Size (height and width) of the maze.
        ambient_size: Size of the final maze matrix. This can be larger than
            `size` to add some visible wall border around the maze. If None, no
            wall border around the maze is produced.
    """
    maze = np.ones((size, size))

    # Start from a random point and recursively open points
    closed_neighbors = []  # Closed points that are neighbors of open points
    
    def _open_point(point):
        # Open a point and add its neighbors to closed_neighbors
        for p in _get_neighbors(size, point):
            if maze[p[0], p[1]] and p not in closed_neighbors:
                closed_neighbors.append(p)
        maze[point[0], point[1]] = 0

    def _find_and_open_new_point():
        # Find a closed neighbor that can be opened without creating an open
        # block, open it, and return True. If no such point exists, return
        # False.
        np.random.shuffle(closed_neighbors)
        for new_point in closed_neighbors:
            if not maze[new_point[0], new_point[1]]:
                continue
            will_make_open_block = any([
                np.sum(maze[i: i + 2, j: j + 2]) <= 1
                for i, j in _get_containing_blocks(size, new_point)
            ])
            if not will_make_open_block:
                _open_point(new_point)
                return True
        return False

    # Seed the maze and iteratively open points
    _open_point(tuple(np.random.randint(0, size, size=(2,))))
    points_to_add = True
    while points_to_add:
        points_to_add = _find_and_open_new_point()

    # Remove dead ends
    _remove_dead_ends(maze)
    
    # If maze has no open points, recurse to generate a new one
    if np.sum(1 - maze) == 0:
        return generate_random_maze_matrix(size, ambient_size=ambient_size)

    # Add wall border if necessary
    if ambient_size is not None and ambient_size > size:
        maze_with_border = np.ones((ambient_size, ambient_size))
        start_index = (ambient_size - size) // 2
        maze_with_border[start_index: start_index + size,
                         start_index: start_index + size] = maze
        maze = maze_with_border

    return maze


def _generate_open_blob(maze, num_points):
    """Try to generate an open connected blob of points in the maze.
    
    Args:
        maze: Instance of .maze.Maze.
        num_points: Int. Number of connected points to have in the blob.

    Returns:
        blob_matrix: False or binary matrix of same size as maze. Ones
            correspond to points in the connected open blob. If False, then
            could not generate a valid blob.
    """
    neighbor_dict = maze.get_neighbor_dict()

    # Seed the blob with a starting point
    blob = [maze.sample_open_point()]
    
    def _get_candidate_new_blob_point():
        # New potential new blob point from neighbors of existing blob point
        candidate_root = blob[np.random.randint(len(blob))]  #pylint: disable=invalid-sequence-index
        neighbors = neighbor_dict[candidate_root]
        candidate = neighbors[np.random.randint(len(neighbors))]
        return candidate
    
    def _add_point():
        # Add a new point to the blob, returning True/False depending on whether
        # this was successful.
        valid_candidate = False
        count = 0
        while not valid_candidate:
            count += 1
            candidate = _get_candidate_new_blob_point()
            if not candidate or candidate in blob:
                continue
            else:
                valid_candidate = True

            if count > _MAX_ITERS:
                return False
        blob.append(candidate)
        return True

    # Add num_points points to the blob if possible, else return False.
    for _ in range(num_points - 1):
        able_to_add_point = _add_point()
        if not able_to_add_point:
            return False

    # Convert the list of blob points to a matrix
    blob_matrix = np.zeros_like(maze.maze)
    for (i, j) in blob:
        blob_matrix[i, j] = 1

    return blob_matrix


def get_connected_open_blob(maze, num_points):
    """Generate an open connected blob of `num_points` points in the maze.
    
    Args:
        maze: Instance of .maze.Maze.
        num_points: Int. Number of connected points to have in the blob.

    Returns:
        blob_matrix: Binary matrix of same size as maze. Ones correspond to
            points in the connected open blob.
    """
    valid_blob = False
    count = 0
    while not valid_blob:
        count += 1
        if count > _MAX_ITERS:
            raise ValueError('Could not generate an open connected blob.')
        
        blob_matrix = _generate_open_blob(maze, num_points)
        if not isinstance(blob_matrix, bool):
            valid_blob = True
    
    return blob_matrix



""".. include:: README.md"""

from .abstract_action_space import AbstractActionSpace
from .composite import Composite
from .grid import Grid
from .joystick import Joystick
from .set_position import SetPosition



# This file was forked and modified from the file here:
# https://github.com/deepmind/spriteworld/blob/master/spriteworld/renderers/pil_renderer.py
# Here is the license header for that file:

# Copyright 2019 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Python Image Library (PIL/Pillow) renderer."""

from . import abstract_observer
from . import color_maps
from . import polygon_modifiers

from dm_env import specs
import numpy as np
from PIL import Image
from PIL import ImageDraw


class PILRenderer(abstract_observer.AbstractObserver):
    """Render using Python Image Library (PIL/Pillow).
    
    This renders an environment state as an image.
    """

    def __init__(self,
                 image_size=(64, 64),
                 anti_aliasing=1,
                 bg_color=None,
                 color_to_rgb=None,
                 polygon_modifier=None):
        """Construct PIL renderer.

        Args:
            image_size: Int tuple (height, width). Size of output of .render().
            anti_aliasing: Int. Anti-aliasing factor. Linearly scales the size
                of the internal canvas.
            bg_color: None or 3-tuple of ints in [0, 255]. Background color. If
                None, background is (0, 0, 0).
            color_to_rgb: String or Callable converting a tuple (c1, c2, c3) to
                a uint8 tuple (r, g, b) in [0, 255]. If string, must be the name
                of a function in color_maps.py, which will be looked up and
                used.
            polygon_modifier: Instance of
                polygon_modifiers.AbstractPolygonModifier. Callable taking state
                and returning a function converting list of polygons (sprite
                vertex arrays) to another list of polygon vertices. This can be
                used to adjust polygon positions to render first-person,
                duplicate sprites when simulating torus geometry, etc. See
                .polygon_modifiers.py for examples.
        """
        self._image_size = image_size
        self._anti_aliasing = anti_aliasing
        self._canvas_size = (anti_aliasing * image_size[0],
                             anti_aliasing * image_size[1])
        
        if polygon_modifier is None:
            polygon_modifier = polygon_modifiers.DoNothing()
        self._polygon_modifier = polygon_modifier

        if color_to_rgb is None:
            color_to_rgb = lambda x: x
        elif isinstance(color_to_rgb, str):
            color_to_rgb = getattr(color_maps, color_to_rgb)
        self.color_to_rgb = color_to_rgb

        self._observation_spec = specs.Array(
            shape=self._image_size + (3,), dtype=np.uint8)

        if bg_color is None:
            bg_color = (0, 0, 0)
        self._canvas_bg = Image.new('RGB', self._canvas_size, bg_color)

        self._canvas = Image.new('RGB', self._canvas_size)
        self._draw = ImageDraw.Draw(self._canvas, 'RGBA')

    def __call__(self, state):
        """Render sprites.

        The order of layers in the state is background to foreground, and the
        order of sprites within layers is also background to foreground.

        Args:
            state: OrderedDict of iterables of sprites.

        Returns:
            Numpy uint8 RGB array of size self._image_size + (3,).
        """
        self._canvas.paste(self._canvas_bg)
        
        polygon_modifier = self._polygon_modifier(state)
        
        for layer in state:
            for sprite in state[layer]:
                polygons = polygon_modifier(layer, sprite)
                for (vertices, color, opacity) in polygons:
                    vertices = self._canvas_size * vertices
                    color = self.color_to_rgb(color)
                    color = tuple(list(color) + [opacity])
                    self._draw.polygon([tuple(v) for v in vertices], fill=color)
        image = self._canvas.resize(self._image_size, resample=Image.LANCZOS)

        # PIL uses a coordinate system with the origin (0, 0) at the upper-left,
        # but our environment uses an origin at the bottom-left (i.e.
        # mathematical convention). Hence we need to flip the render vertically
        # to correct for that.
        image = np.flipud(np.array(image))
        
        return image

    @property
    def polygon_modifier(self):
        return self._polygon_modifier

    def observation_spec(self):
        return self._observation_spec



"""Rigid tether physics.

The two classes Tether and TetherZippedLayers can be used to rigidly tether
together sets of sprites, optionally also tethering them to an anchor point in
the environment.

These classes are typically used as corrective physics for a physics.Physics
instance.
"""

from . import abstract_physics
import itertools
import numpy as np


def _change_rotation_coordinates(sprite,
                                 origin,
                                 origin_velocity,
                                 updates_per_env_step):
    """Change rotation coordinates to new origin.
    
    This function gets the angular momentum, moment of inertia, radius, and
    perpendicular relative to a new origin point.
    """
    # Get angular momentum coming from velocity
    delta_position = sprite.velocity / updates_per_env_step
    parallel = (sprite.position + 0.5 * delta_position) - origin
    radius = np.linalg.norm(parallel)
    parallel /= radius
    perpendicular = np.matmul(np.array([[0, -1], [1, 0]]), parallel)
    perp_vel = np.dot(sprite.velocity - origin_velocity, perpendicular)
    angular_momentum = perp_vel * sprite.mass * radius

    # Add angular momentum coming from angular velocity
    angular_momentum += sprite.angle_vel * sprite.moment_of_inertia

    # Get moment of inertia
    moment_of_inertia = sprite.moment_of_inertia + sprite.mass * radius * radius

    return angular_momentum, moment_of_inertia, radius, perpendicular


def _tether_sprites(sprites,
                    updates_per_env_step,
                    update_angle_vel=True,
                    anchor=None):
    """Apply a tether to a set of sprites."""
    # Can't tether if no sprites
    if len(sprites) == 0:
        return
        
    # Compute center of mass
    total_mass = sum([s.mass for s in sprites])
    
    # Can't tether if infinite masses are involved
    if np.isinf(total_mass):
        return
    
    center_of_mass = (
        sum([s.mass * s.position for s in sprites]) / total_mass)

    # Compute total momentum and velocity
    total_momentum = sum([s.mass * s.velocity for s in sprites])
    total_velocity = total_momentum / total_mass

    if anchor is not None:
        center_of_mass = anchor
        total_velocity = np.zeros(2)

    if update_angle_vel:
        # Compute total angular momentum and angular velocity
        angular_momenta, moments_of_inertia, radii, perpendiculars = zip(*[
            _change_rotation_coordinates(
                s, center_of_mass, total_velocity, updates_per_env_step)
            for s in sprites
        ])
        total_angular_momentum = sum(angular_momenta)
        total_moment_of_inertia = sum(moments_of_inertia)
        total_angular_velocity = (
            total_angular_momentum / total_moment_of_inertia)

        # Set velocities and angular velocities
        for s, radius, perp in zip(sprites, radii, perpendiculars):
            s.velocity = (
                total_velocity + radius * perp * total_angular_velocity)
            s.angle_vel = total_angular_velocity
    else:
        # Set velocities and angular velocities
        for s in sprites:
            s.velocity = total_velocity
            s.angle_vel = 0.


class Tether(abstract_physics.AbstractPhysics):
    """Rigid tether physics class.
    
    This is used to rigidly tether all sprites in specified layers. For example,
    if you want all prey to be rigidly connected, use Tether('prey'). It is
    typically used as a corrective_physics argument to physics.Physics.

    WARNING: This rigid tether does not work if sprites have their positions
    changed directly --- it only works to adjust sprite velocities and angular
    velocities. This can sometimes have noticeable effects if tethered sprites
    have some collisions.
    """

    def __init__(self, layer_names, update_angle_vel=True, anchor=None):
        """Constructor.
        
        Args:
            layer_names: String or iterable of strings. All elements must be
                keys in environment state. All sprites in all of these layers
                will be tethered and will move together.
            update_angle_vel: Bool. If True, fully simulate the rotational
                mechanics and update the sprites angular velocity.
            anchor: Optional anchor point. If given, sprites will be tethered to
                this fixed point and will remain at a fixed distance from it.
        """
        if not isinstance(layer_names, (list, tuple)):
            self._layer_names = [layer_names]
        else:
            self._layer_names = layer_names
        self._update_angle_vel = update_angle_vel
        self._anchor = anchor

    def apply_physics(self, state, updates_per_env_step):
        """Step the physics.
        
        Args:
            sprites: Environment state. Dictionary of iterables of instances of
                ../sprite.Sprite.
            updates_per_env_step: Int. Number of times this physics is called
                per environment step. Unused in this method, but needed to
                satisfy the AbstractPhysics signature.
        """
        sprites = list(itertools.chain(
            *[state[layer_name] for layer_name in self._layer_names]))
        _tether_sprites(
            sprites, updates_per_env_step,
            update_angle_vel=self._update_angle_vel, anchor=self._anchor)


class TetherZippedLayers(abstract_physics.AbstractPhysics):
    """Apply rigid tethers between zipped sprites across layers.
    
    This zips the sprites in different layers and tethers them together.
    Specifically, for each index i, the set {i'th sprite in each layer} is
    tethered together, assuming all provided layers have the same number of
    sprites.
    
    The is useful for example if you want each sprite in a layer to carry a
    local occluder sprite with it, as in spatial delayed match-to-sample.

    WARNING: As with Tether above, this class does not work if sprites have
    their positions changed directly --- it only works to adjust sprite
    velocities and angular velocities.
    """

    def __init__(self, layer_names, update_angle_vel=True, anchor=None):
        """Constructor.
        
        Args:
            layer_names: String or iterable of strings. All elements must be
                keys in environment state. Sprites these layers will be zipped
                and tethered. These layers must all have the same number of
                sprites.
            update_angle_vel: Bool. If True, fully simulate the rotational
                mechanics and update the sprites angular velocity.
            anchor: Optional anchor point. If given, sprites will be tethered to
                this fixed point and will remain at a fixed distance from it.
        """
        if not isinstance(layer_names, (list, tuple)):
            self._layer_names = [layer_names]
        else:
            self._layer_names = layer_names
        self._update_angle_vel = update_angle_vel
        self._anchor = anchor

    def apply_physics(self, state, updates_per_env_step):
        """Step the physics.
        
        Args:
            sprites: Environment state. Dictionary of iterables of instances of
                ../sprite.Sprite.
            updates_per_env_step: Int. Number of times this physics is called
                per environment step. Unused in this method, but needed to
                satisfy the AbstractPhysics signature.
        """
        # Ensure that all layers have the same number of sprites
        layer_sprites = [state[layer_name] for layer_name in self._layer_names]
        layer_lengths = [len(x) for x in layer_sprites]
        if not all(length == layer_lengths[0] for length in layer_lengths):
            raise ValueError(
                'All layers fed into TetherAcrossLayers must have the same '
                'number of sprites, but their counts are {}'.format(
                    layer_lengths))
        
        for sprites in zip(*layer_sprites):
            _tether_sprites(
                sprites, updates_per_env_step,
                update_angle_vel=self._update_angle_vel, anchor=self._anchor)



"""Abstract wrapper.

This file contains AbstractEnvironmentWrapper, an abstract base class for
environment wrappers that mimics the interface of the underlying environment.
"""

import abc


class AbstractEnvironmentWrapper(abc.ABC):
    """Abstract environment wrapper class.
    
    All environment wrappers must inherit from this class.
    """

    def __init__(self, environment):
        self._environment = environment

    def reset(self):
        return self._environment.reset()
    
    def step(self, action):
        return self._environment.step(action)

    def observation(self):
        return self._environment.observation()

    def observation_spec(self):
        return self._environment.observation_spec()

    def action_spec(self):
        return self._environment.action_spec()
    
    @property
    def state(self):
        return self._environment.state

    @property
    def meta_state(self):
        return self._environment.meta_state

    @property
    def state_initializer(self):
        return self._environment.state_initializer

    @property
    def physics(self):
        return self._environment.physics

    @property
    def task(self):
        return self._environment.task

    @property
    def action_space(self):
        return self._environment.action_space

    @property
    def observers(self):
        return self._environment.observers

    @property
    def game_rules(self):
        return self._environment.game_rules
    
    @property
    def environment(self):
        return self._environment
    
    @property
    def step_count(self):
        return self._environment.step_count

    @property
    def reset_next_step(self):
        return self._environment.reset_next_step



# This file was forked and modified from the file here:
# https://github.com/deepmind/spriteworld/blob/master/spriteworld/factor_distributions.py
# Here is the license header for that file:

# Copyright 2019 DeepMind Technologies Limited.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ============================================================================
"""Factor distribution library.

This library contains classes for defining distributions of sprite factors.
A number of set-theoretic operations are supported, with which it is possible to
define factor distributions that are arbitrarily nested mixtures, intersections,
products, and differences of single-factor continuous/discrete distributions.

A factor specification is called a "spec", which is a dictionary of sprite
factors, hence can have keys such as "size", "shape", "x_pos", etc. However, the
classes in this file are general and make no reference to the particular factor
names used by Spriteworld sprites.

All distributions inherit from AbstractDistribution. They have a "sample()"
method, which returns a spec. The keys of this spec can be accessed by the
"keys" property. Distributions also have a "contains(spec)" method, which checks
if the argument "spec" is in the support of the distribution.
"""

import abc
import functools
import numpy as np

# Maximum number of tries used for rejection sampling from Intersection and
# SetMinus distributions
_MAX_TRIES = int(1e5)


class AbstractDistribution(abc.ABC):
    """Abstract class from which all distributions should inherit."""

    @abc.abstractmethod
    def sample(self, rng=None):
        """Sample a spec from this distribution. Returns a dictionary.
        
        Args:
            rng: Random number generator. Fed into self._get_rng(), if None
                defaults to np.random.
        """

    @abc.abstractmethod
    def contains(self, spec):
        """Return whether distribution contains spec dictionary."""

    @abc.abstractmethod
    def to_str(self, indent):
        """Recursive string description of this distribution."""

    def __str__(self):
        return self.to_str(indent=0)

    def _get_rng(self, rng=None):
        """Get random number generator, defaulting to np.random."""
        return np.random if rng is None else rng

    @abc.abstractproperty
    def keys(self):
        """The set of keys in specs sampled from this distribution."""


class Continuous(AbstractDistribution):
    """Continuous 1-dimensional uniform distribution."""

    def __init__(self, key, minval, maxval, dtype='float32'):
        """Construct continuous 1-dimensional uniform distribution.
        
        Args:
            key: String factor name. self.sample() returns {key: _}.
            minval: Scalar minimum value.
            maxval: Scalar maximum value.
            dtype: String numpy dtype.
        """
        self.key = key
        self.minval = minval
        self.maxval = maxval
        self.dtype = dtype

    def sample(self, rng=None):
        """Sample value in [self.minval, self.maxval) and return dict."""
        rng = self._get_rng(rng)
        out = rng.uniform(low=self.minval, high=self.maxval)
        out = np.asarray(out, dtype=self.dtype)
        return {self.key: out}

    def contains(self, spec):
        """Check if spec[self.key] is in [self.minval, self.maxval)."""
        if self.key not in spec:
            raise KeyError('key {} is not in spec {}, but must be to evaluate '
                           'containment.'.format(self.key, spec))
        else:
            return (
                spec[self.key] >= self.minval and spec[self.key] < self.maxval)

    def to_str(self, indent):
        s = '<Continuous: key={}, mival={}, maxval={}, dtype={}>'.format(
            self.key, self.minval, self.maxval, self.dtype)
        return indent * '  ' + s

    @property
    def keys(self):
        return set([self.key])


class Discrete(AbstractDistribution):
    """Discrete distribution."""

    def __init__(self, key, candidates, probs=None):
        """Construct discrete distribution.
        
        Args:
            key: String. Factor name.
            candidates: Iterable. Discrete values to sample from.
            probs: None or iterable of floats summing to 1. Candidate sampling
                probabilities. If None, candidates are sampled uniformly.
        """
        self.candidates = candidates
        self.key = key
        self.probs = probs

    def sample(self, rng=None):
        rng = self._get_rng(rng)
        out = self.candidates[rng.choice(len(self.candidates), p=self.probs)]
        return {self.key: out}

    def contains(self, spec):
        if self.key not in spec:
            raise KeyError('key {} is not in spec {}, but must be to evaluate '
                           'containment.'.format(self.key, spec))
        else:
            return spec[self.key] in self.candidates

    def to_str(self, indent):
        s = '<Discrete: key={}, candidates={}, probs={}>'.format(
            self.key, self.candidates, self.probs)
        return indent * '  ' + s

    @property
    def keys(self):
        return set([self.key])


class Mixture(AbstractDistribution):
    """Mixture of distributions."""

    def __init__(self, components, probs=None):
        """Construct mixture of distributions.
        
        This is a mixture distribution, not a union, so if the components
        overlap, their overlap will be sampled more than the non-overlapping
        regions.
        
        Args:
            components: Iterable of component distributions. Must all have the
                same key sets.
            probs: None or iterable of floats summing to 1. Sampling
                probabilities for the components.
        """
        self.components = components
        if probs is None:
            self.probs = np.ones(len(components)) / len(components)
        else:
            self.probs = np.array(probs)

        self._keys = components[0].keys
        for c in components[1:]:
            if c.keys != self._keys:
                raise ValueError(
                    'All components must have the same key sets. However '
                    'detected key sets {} and {}'.format(self._keys, c.keys))

    def sample(self, rng=None):
        rng = self._get_rng(rng)
        sample_index = rng.choice(len(self.components), p=self.probs)
        sample = self.components[sample_index].sample(rng=rng)
        return sample

    def contains(self, spec):
        return any(c.contains(spec) for c in self.components)

    def to_str(self, indent):
        components_strings = [x.to_str(indent + 2) for x in self.components]
        s = (indent * '  ' + '<Mixture:\n' +
            (indent + 1) * '  ' + 'components=[\n{},\n' +
            (indent + 1) * '  ' + '],\n' +
            (indent + 1) * '  ' + 'probs={}>').format(
                ',\n'.join(components_strings), self.probs)
        return s

    @property
    def keys(self):
        return self._keys


class Intersection(AbstractDistribution):
    """Intersection of component distributions."""

    def __init__(self, components, index_for_sampling=0):
        """Construct intersection of component distributions.
        
        Samples are generated by sampling from one of the components and then
        doing rejection with the others, so if the component being sampled has
        some non-uniformity (e.g. a mixture with non-uniform probs), that
        non-uniformity will be inherited by the intersection.
        
        Args:
            components: Iterable of distributions.
            index_for_sampling: Int. Index of the component to use for sampling.
                All other components will be used to reject its samples. For
                efficiency, the user should ensure index_for_sampling
                corresponds to the smallest component distribution.
        """
        self.components = components
        self.index_for_sampling = index_for_sampling

        self._keys = components[0].keys
        for c in components[1:]:
            if c.keys != self._keys:
                raise ValueError(
                    'All components must have the same key sets. However '
                    'detected key sets {} and {}'.format(self._keys, c.keys))

    def sample(self, rng=None):
        rng = self._get_rng(rng)
        tries = 0
        while tries < _MAX_TRIES:
            tries += 1
            sample = self.components[self.index_for_sampling].sample(rng=rng)
            if all(c.contains(sample) for c in self.components):
                return sample
        raise ValueError('Maximum number of tried exceeded when trying to '
                         'sample from {}.'.format(str(self)))

    def contains(self, spec):
        return all(c.contains(spec) for c in self.components)

    def to_str(self, indent):
        components_strings = [x.to_str(indent + 2) for x in self.components]
        s = (indent * '  ' + '<Intersection:\n' +
            (indent + 1) * '  ' + 'components=[\n{},\n' +
            (indent + 1) * '  ' + '],\n' +
            (indent + 1) * '  ' + 'index_for_sampling={}>').format(
                ',\n'.join(components_strings), self.index_for_sampling)
        return s

    @property
    def keys(self):
        return self._keys


class Product(AbstractDistribution):
    """Product distribution."""

    def __init__(self, components, **constants):
        """Construct product distribution.
        
        This is used to create distributions over larger numbers of factors by
        taking the product of components. The components must have disjoint key
        sets.
        
        Args:
            components: Iterable of distributions.
            constants: Dictionary. Keys will be additional factors to the
                distribution and values will the constant values for those keys.
                So using constants is an easy way to effectively pass in extra
                Discrete 1-candidate distributions.
        """
        constant_components = [Discrete(k, [v]) for k, v in constants.items()]
        components = list(components) + constant_components
        self.components = components

        self._keys = functools.reduce(
            set.union, [set(c.keys) for c in components])
        num_keys = sum(len(c.keys) for c in components)
        if len(self._keys) < num_keys:
            raise ValueError(
                'All components must have different keys, yet there are {} '
                'overlapping keys.'.format(num_keys - len(self._keys)))

    def sample(self, rng=None):
        rng = self._get_rng(rng)
        sample = {}
        for c in self.components:
            sample.update(c.sample(rng=rng))
        return sample

    def contains(self, spec):
        return all(c.contains(spec) for c in self.components)

    def to_str(self, indent):
        components_strings = [x.to_str(indent + 2) for x in self.components]
        s = (indent * '  ' + '<Product:\n' +
            (indent + 1) * '  ' + 'components=[\n{},\n' +
            (indent + 1) * '  ' + ']>').format(
                ',\n'.join(components_strings))
        return s

    @property
    def keys(self):
        return self._keys


class SetMinus(AbstractDistribution):
    """Setminus of distributions."""

    def __init__(self, base, hold_out):
        """Construct setminus of distributions..
        
        This uses rejection sampling to take the difference of two
        distributions.
        
        Args:
            base: Distribution from which candidate samples are drawn.
            hold_out: Distribution used to reject samples from base.
        """
        self.base = base
        self.hold_out = hold_out

        self._keys = base.keys
        if not hold_out.keys.issubset(self._keys):
            raise ValueError(
                'Keys {} of hold_out is not a subset of keys {} of SetMinus '
                'base distribution.'.format(hold_out.keys, base.keys))

    def sample(self, rng=None):
        rng = self._get_rng(rng)
        tries = 0
        while tries < _MAX_TRIES:
            tries += 1
            sample = self.base.sample(rng=rng)
            if not self.hold_out.contains(sample):
                return sample
        raise ValueError('Maximum number of tried exceeded when trying to '
                         'sample from {}.'.format(str(self)))

    def contains(self, spec):
        return self.base.contains(spec) and not self.hold_out.contains(spec)

    def to_str(self, indent):
        s = (indent * '  ' + '<SetMinus:\n' +
            (indent + 1) * '  ' + 'base=\n{},\n' +
            (indent + 1) * '  ' + 'hold_out=\n{}>').format(
                self.base.to_str(indent + 2), self.hold_out.to_str(indent + 2))
        return s

    @property
    def keys(self):
        return self._keys


class Selection(AbstractDistribution):
    """Filter a source distribution."""

    def __init__(self, base, filtering):
        """Construct selection of a base distribution given a filter.
        
        Given a base Distribution and a filter Distribution, returns samples of
        the base which are compatible with the filter.
        
        This is related to Intersection, but does not expect the base and
        filters to have the same keys. Instead, the filters should be subsets of
        the base. This is the same as SetMinus, except the filter accepts
        instead of rejects samples.

        Args:
            base: Distribution from which candidate samples are drawn.
            filtering: Distribution used to select samples from base.
        """
        self.base = base
        self.filtering = filtering

        self._keys = base.keys
        if not filtering.keys.issubset(self._keys):
            raise ValueError(
                'Keys {} of filtering is not a subset of keys {} of Selection '
                'base distribution.'.format(filtering.keys, base.keys))

    def sample(self, rng=None):
        rng = self._get_rng(rng)
        tries = 0
        while tries < _MAX_TRIES:
            tries += 1
            sample = self.base.sample(rng=rng)
            if self.filtering.contains(sample):
                return sample
        raise ValueError(
            'Maximum number of tried exceeded when trying to sample from {}.'
            .format(str(self)))

    def contains(self, spec):
        return self.base.contains(spec) and self.filtering.contains(spec)

    def to_str(self, indent):
        s = (indent * '  ' + '<Selection:\n' + (indent + 1) * '  ' +
            'base=\n{},\n' + (indent + 1) * '  ' + 'filtering=\n{}>').format(
                self.base.to_str(indent + 2), self.filtering.to_str(indent + 2))
        return s

    @property
    def keys(self):
        return self._keys


class DependentDistribution(AbstractDistribution):
    """Distibution in which some factors depend deterministically on others.
    
    For example, suppose you want a distribution over keys ['x', 'y'] where
    values are floats in [0, 1] but y = 1 - x. This could be done with:
    ```python
        DependentDistribution(
            independent_distrib=Continuous('x', 0., 1.),
            dependent_fn=lambda indep_sample: {'y': 1. - indep_sample['x']},
            dependent_fn_keys=['y'],
        )
    ```
    """

    def __init__(self, independent_distrib, dependent_fn, dependent_fn_keys):
        """Constructor.

        Args:
            independent_distrib: Instance of AbstractDistribution.
            dependent_fn: Function taking a sample from independent_distrib and
                returning a dictionary.
            dependent_fn_keys: Iterable of keys of the output of dependent_fn.
        """
        self._independent_distrib = independent_distrib
        self._dependent_fn = dependent_fn
        self._dependent_fn_keys = dependent_fn_keys

        if not set(independent_distrib.keys).isdisjoint(set(dependent_fn_keys)):
            raise ValueError(
                'independent_distrib keys {} and dependent_fn keys {} are not '
                'disjoint.'.format(independent_distrib.keys, dependent_fn_keys))

    def sample(self, rng=None):
        rng = self._get_rng(rng)
        sample = self._independent_distrib.sample(rng=rng)
        sample.update(self._dependent_fn(sample))
        return sample

    def contains(self, spec):
        contains = self._independent_distrib.contains(spec)
        sub_spec = {k: spec[k] for k in self._independent_distrib.keys}
        dependent_fn_sub_spec = self._dependent_fn(sub_spec)
        for k in self._dependent_fn_keys:
            contains &= spec[k] == dependent_fn_sub_spec[k]
        return contains

    @property
    def keys(self):
        return self._independent_distrib.keys.union(self._dependent_fn_keys)

    def to_str(self, indent):
        s = (indent * '  ' + '<DependentDistribution:\n' +
            (indent + 1) * '  ' + 'independent_distrib=\n{},\n' +
            (indent + 1) * '  ' + 'dependent_fn={}>').format(
                self._independent_distrib, self._dependent_fn)
        return s



"""SetPosition action space for controlling agent avatars.

This action space sets the position of an agent. This is commonly used with a
touch-screen or mouse-click interface.
"""

from . import abstract_action_space
from dm_env import specs
import numpy as np


class SetPosition(abstract_action_space.AbstractActionSpace):
    """SetPosition action space."""

    def __init__(self, action_layers='agent', inertia=0.):
        """Constructor.
        
        Args:
            agent_layer: String or iterable of strings. Elements (or itself if
                string) must be keys in the environment state. All sprites in
                these layers will be acted upon by this action space.
            inertia: Float in [0, 1]. Inertia of the action layer sprite
                positions. Zero means position is set instantly to action value,
                and 1 means the action has no effect on sprite positions.
        """
        if not isinstance(action_layers, (list, tuple)):
            action_layers = (action_layers,)
        self._action_layers = action_layers
        self._inertia = inertia

        self._action_spec = specs.BoundedArray(
            shape=(2,), dtype=np.float32, minimum=0, maximum=1)

    def step(self, state, action):
        """Apply action to environment state.

        Args:
            state: Ordereddict. Environment state.
            action: Numpy float array of size (2), in [0, 1]. Position to set
                for the agent(s) in self._action_layers.
        """
        for action_layer in self._action_layers:
            for sprite in state[action_layer]:
                sprite.position = (
                    self._inertia * sprite.position +
                    (1 - self._inertia) * action
                )

    def random_action(self):
        """Return randomly sampled action."""
        return np.random.uniform(0., 1., size=(2,))
            
    def action_spec(self):
        return self._action_spec



"""Task that gives reward at regular intervals during task."""

from . import abstract_task


class StayAlive(abstract_task.AbstractTask):
    """StayAlive task. In this task a reward is given at regular intervals."""

    def __init__(self, reward_period, reward_value=1.):
        """Constructor.

        Args:
            reward_period: Int. Number of steps between each reward.
            reward_value: Scalar. Value of reward given.
        """
        self._reward_period = reward_period
        self._reward_value = reward_value

    def reset(self, state, meta_state):
        pass

    def reward(self, state, meta_state, step_count):
        """Compute reward."""
        del state
        del meta_state
        
        if (step_count + 1) % self._reward_period == 0:
            reward = self._reward_value
        else:
            reward = 0
            
        return reward, False



"""Physics class.

The `physics` argument to ..environment.Environment is typically an instance of
the Physics class in this file.
"""

from . import abstract_physics
import itertools
import numpy as np


class Physics(abstract_physics.AbstractPhysics):
    """Force physics class."""

    def __init__(self, *forces, updates_per_env_step=1, corrective_physics=()):
        """Constructor.

        Examples of usage:

            ```python
            friction_force = Friction(coeff_friction=1)
            spring_force = Spring(equilibrium=0.3, spring_const=0.1)
            collision_force = Collision(elasticity=0.5)

            # Specify only friction on the avatar sprite(s)
            force = (friction_force, 'avatar')
            my_physics = Physics(force)

            # Specify friction on the avatar and prey sprite(s)
            force = (friction_force, ['avatar', 'prey'])
            my_physics = Physics(force)

            # Specify friction on avatar and springs between avatar and all prey
            forces = (
                (friction_force, 'avatar'),
                (spring_force, 'avatar', 'prey'),
            )
            my_physics = Physics(*forces)

            # Add collisions between all avatar/prey sprites and walls/neutrals
            forces = (
                (friction_force, 'avatar'),
                (spring_force, 'avatar', 'prey'),
                (collision_force, ['avatar', 'prey'], ['walls', 'neutrals']),
            )
            my_physics = Physics(*forces)
            ```

        Args:
            *forces: Iterable. Each element is a tuple (force_instance, *args)
                where force_instance is an instance of
                abstract_force.AbstractForce and *args has the same length as
                the number of arguments of force_instance's step() method. Each
                element of args may be either a string key of the environment
                state or an iterable of such. This specifies all sprites to feed
                into the force.
            updates_per_env_step: Int. Number of physics applications per step
                of the environment. A higher value means smoother physics but
                slower runtime. This is fed into the individual forces so that
                they can accomodate their force strength accordingly.
            corrective_physics: Optional instance (or iterable of instances) of
                abstract_physics.AbstractPhysics to be applied every step before
                updating the sprite positions. This is typically used to apply
                corrections to sprite velocities. For example, it can be used to
                enforce rigid tethers between sprites. 
        """
        super(Physics, self).__init__(updates_per_env_step=updates_per_env_step)
        self._forces = forces
        if not isinstance(corrective_physics, (list, tuple)):
            corrective_physics = [corrective_physics]
        self._corrective_physics = corrective_physics
    
    def reset(self, state):
        for force in self._forces:
            force[0].reset(state)
        
        for corrective_physics in self._corrective_physics:
            corrective_physics.reset(state)

    def _args_to_iterables(self, args_iterables):
        """Converts all non-iterable elements to singleton tuples."""
        args_iterables = [
            x if isinstance(x, (tuple, list)) else (x,)
            for x in args_iterables
        ]
        return args_iterables

    def apply_physics(self, state, updates_per_env_step):
        """Move the sprites according to the physics."""

        # Update sprites based on forces
        for (force, *args_iterables) in self._forces:
            args_iterables = self._args_to_iterables(args_iterables)

            # If args_iterables = [['avatar', 'prey'], ['walls', 'neutrals']],
            # then
            # args_combinations = [
            #   ['avatar', 'walls'],
            #   ['avatar', 'neutrals'],
            #   ['prey', 'walls'],
            #   ['prey', 'neutrals'],
            # ]
            args_combinations = itertools.product(*args_iterables)

            for args in args_combinations:
                for sprites in itertools.product(*[state[s] for s in args]):
                    force.step(
                        *sprites, updates_per_env_step=updates_per_env_step)

        for corrective_physics in self._corrective_physics:
            corrective_physics.apply_physics(state, updates_per_env_step)

        # Move sprites based on their velocity
        delta_t = 1. / updates_per_env_step
        for layer in state:
            for sprite in state[layer]:
                sprite.update_pos_from_vel(delta_t=delta_t)



"""PacMan-like task.

In this task all objects move in a maze with the same speed. The maze is
randomized every trial. The subject controls a green agent. Red ghost agents
wander the maze randomly without backtracking. The subject's goal is to collect
all yellow pellets in the maze. Ghosts only begin moving once agent moves.
"""

import collections
import numpy as np

from moog import action_spaces
from moog import game_rules
from moog import maze_lib
from moog import observers
from moog import physics as physics_lib
from moog import sprite
from moog import tasks


def _get_config(num_ghosts, maze_size):
    """Get environment config."""

    ############################################################################
    # Sprite initialization
    ############################################################################

    # Agent
    agent_factors = dict(shape='circle', scale=0.05, c0=0.33, c1=1., c2=0.66)

    # Prey
    prey_factors = dict(shape='circle', scale=0.025, c0=0.2, c1=1., c2=1.)

    # Ghosts
    ghost_factors = dict(
        shape='circle', scale=0.05, mass=np.inf, c0=0., c1=1., c2=0.8)

    def state_initializer():
        maze = maze_lib.generate_random_maze_matrix(
            size=maze_size, ambient_size=12)
        maze = maze_lib.Maze(np.flip(maze, axis=0))
        walls = maze.to_sprites(c0=0., c1=0., c2=0.8)

        # Sample positions in maze grid of agent and ghosts
        n_ghosts = num_ghosts()
        points = maze.sample_distinct_open_points(1 + n_ghosts)
        positions = [maze.grid_side * (0.5 + np.array(x)) for x in points]

        # Agent
        agent_position = positions[0]
        agent = [sprite.Sprite(
            x=agent_position[1], y=agent_position[0], **agent_factors)]

        # ghosts
        ghosts = []
        for i in range(n_ghosts):
            position = positions[i + 1]
            ghosts.append(sprite.Sprite(
                x=position[1], y=position[0], **ghost_factors))

        # Place prey at every open maze location
        prey = []
        open_maze_points = np.argwhere(maze.maze == 0)
        for p in open_maze_points:
            pos = maze.grid_side * (0.5 + np.array(p))
            prey.append(sprite.Sprite(x=pos[1], y=pos[0], **prey_factors))

        state = collections.OrderedDict([
            ('walls', walls),
            ('prey', prey),
            ('ghosts', ghosts),
            ('agent', agent),
        ])
        return state

    ############################################################################
    # Physics
    ############################################################################

    maze_physics = physics_lib.MazePhysics(
        maze_layer='walls',
        avatar_layers=('agent', 'prey', 'ghosts'), constant_speed=0.015,
    )

    physics = physics_lib.Physics(
        (physics_lib.RandomMazeWalk(speed=0.015), ['ghosts']),
        updates_per_env_step=1, corrective_physics=[maze_physics],
    )

    ############################################################################
    # Task
    ############################################################################

    ghost_task = tasks.ContactReward(
        -5, layers_0='agent', layers_1='ghosts', reset_steps_after_contact=0)
    prey_task = tasks.ContactReward(1, layers_0='agent', layers_1='prey')
    reset_task = tasks.Reset(
        condition=lambda state: len(state['prey']) == 0,
        steps_after_condition=5,
    )
    task = tasks.CompositeTask(
        ghost_task, prey_task, reset_task, timeout_steps=1000)

    ############################################################################
    # Action space
    ############################################################################

    action_space = action_spaces.Grid(
        scaling_factor=0.015,
        action_layers='agent',
        control_velocity=True,
        momentum=0.5,  # Value irrelevant, since maze_physics has constant speed
    )

    ############################################################################
    # Observer
    ############################################################################

    observer = observers.PILRenderer(
        image_size=(256, 256),
        anti_aliasing=1,
        color_to_rgb='hsv_to_rgb',
    )

    ############################################################################
    # Game rules
    ############################################################################

    def _unglue(s):
        s.mass = 1.
    def _unglue_condition(state):
        return not np.all(state['agent'][0].velocity == 0)
    unglue = game_rules.ConditionalRule(
        condition=_unglue_condition,
        rules=game_rules.ModifySprites(('prey', 'ghosts'), _unglue),
    )

    vanish_on_contact = game_rules.VanishOnContact(
        vanishing_layer='prey', contacting_layer='agent')

    rules = (vanish_on_contact, unglue)

    ############################################################################
    # Final config
    ############################################################################

    config = {
        'state_initializer': state_initializer,
        'physics': physics,
        'task': task,
        'action_space': action_space,
        'observers': {'image': observer},
        'game_rules': rules,
    }
    return config


def get_config(level):
    """Get config dictionary of kwargs for environment constructor.
    
    Args:
        level: Int. Different values yield different maze sizes and numbers of
            ghosts.
    """
    if level == 0:
        return _get_config(
            num_ghosts=lambda: 2,
            maze_size=8,
        )
    elif level == 1:
        return _get_config(
            num_ghosts=lambda: 3,
            maze_size=10,
        )
    else:
        raise ValueError('Invalid level {}'.format(level))



"""Chase/avoid task on a torus.

In this task there are predators (red circles), prey (yellow circles) and an
agent (green square). The subject controls the agent with a joystick. The
subject's goal is to catch the prey while avoiding being caught by the
predators. The prey are repulsed by the agent and predators are attracted to the
agent. Prey and predators move stochasticity and with constant speed.

The environment geometry is shaped like a torus --- when an object reaches one
boundary, it reappears on the opposite boundary.
"""

import collections
import itertools
import numpy as np

from moog import action_spaces
from moog import game_rules
from moog import observers
from moog import physics as physics_lib
from moog import sprite
from moog import tasks
from moog.state_initialization import distributions as distribs
from moog.state_initialization import sprite_generators
from moog.observers import polygon_modifiers


def _get_config(num_prey, num_predators):
    """Get environment config."""

    ############################################################################
    # Sprite initialization
    ############################################################################

    # Agent
    agent_factors = distribs.Product(
        [distribs.Continuous('x', 0., 1.),
         distribs.Continuous('y', 0., 1.)],
        scale=0.08, c0=0, c1=255, c2=0,
    )

    # Predators
    predator_factors = distribs.Product(
        [distribs.Continuous('x', 0., 1.),
         distribs.Continuous('y', 0., 1.),
         distribs.Continuous('x_vel', -0.02, 0.02),
         distribs.Continuous('y_vel', -0.02, 0.02),],
        scale=0.08, shape='circle', opacity=192, c0=255, c1=0, c2=0,
    )

    # Prey
    prey_factors = distribs.Product(
        [distribs.Continuous('x', 0., 1.),
         distribs.Continuous('y', 0., 1.),
         distribs.Continuous('x_vel', -0.02, 0.02),
         distribs.Continuous('y_vel', -0.02, 0.02),],
        scale=0.08, shape='circle', opacity=192, c0=255, c1=255, c2=0,
    )

    # Create callable initializer returning entire state
    predator_generator = sprite_generators.generate_sprites(
        predator_factors, num_sprites=num_predators)
    prey_generator = sprite_generators.generate_sprites(
        prey_factors, num_sprites=num_prey)

    def state_initializer():
        """Callable returning state at every episode reset."""
        agent = sprite.Sprite(**agent_factors.sample())
        predators = predator_generator(without_overlapping=(agent,))
        prey = prey_generator(without_overlapping=(agent,))

        state = collections.OrderedDict([
            ('prey', prey),
            ('predators', predators),
            ('agent', [agent]),
        ])
        return state

    ############################################################################
    # Physics
    ############################################################################

    agent_friction_force = physics_lib.Drag(coeff_friction=0.25)
    random_force = physics_lib.RandomForce(max_force_magnitude=0.01)
    predator_attraction = physics_lib.DistanceForce(
        physics_lib.linear_force_fn(zero_intercept=-0.001, slope=0.0005))
    prey_avoid = physics_lib.DistanceForce(
        physics_lib.linear_force_fn(zero_intercept=0.001, slope=-0.0005))

    forces = (
        (agent_friction_force, 'agent'),
        (random_force, ['predators', 'prey']),
        (predator_attraction, 'agent', 'predators'),
        (prey_avoid, 'agent', 'prey'),
    )

    constant_speed = physics_lib.ConstantSpeed(
        layer_names=['prey', 'predators'], speed=0.015)

    physics = physics_lib.Physics(
        *forces,
        updates_per_env_step=10,
        corrective_physics=[constant_speed],
    )

    ############################################################################
    # Task
    ############################################################################

    predator_task = tasks.ContactReward(
        -5, layers_0='agent', layers_1='predators', reset_steps_after_contact=0)
    prey_task = tasks.ContactReward(1, layers_0='agent', layers_1='prey')
    reset_task = tasks.Reset(
        condition=lambda state: len(state['prey']) == 0,
        steps_after_condition=5,
    )
    task = tasks.CompositeTask(
        reset_task, predator_task, prey_task, timeout_steps=300)

    ############################################################################
    # Action space
    ############################################################################

    action_space = action_spaces.Joystick(
        scaling_factor=0.025, action_layers='agent', control_velocity=True)

    ############################################################################
    # Observer
    ############################################################################

    observer = observers.PILRenderer(
        image_size=(64, 64),
        anti_aliasing=1,
        polygon_modifier=polygon_modifiers.TorusGeometry(
            ['agent', 'predators', 'prey']),
    )

    ############################################################################
    # Game rules
    ############################################################################

    prey_vanish = game_rules.VanishOnContact(
        vanishing_layer='prey', contacting_layer='agent')
    def _torus_position_wrap(s):
        s.position = np.remainder(s.position, 1)
    torus_position_wrap = game_rules.ModifySprites(
        ('agent', 'predators', 'prey'), _torus_position_wrap)

    rules = (prey_vanish, torus_position_wrap)

    ############################################################################
    # Final config
    ############################################################################

    config = {
        'state_initializer': state_initializer,
        'physics': physics,
        'task': task,
        'action_space': action_space,
        'observers': {'image': observer},
        'game_rules': rules,
    }
    return config


def get_config(level):
    if level == 0:
        return _get_config(
            num_prey=1,
            num_predators=2,
        )
    elif level == 1:
        return _get_config(
            num_prey=lambda: np.random.randint(1, 3),
            num_predators=lambda: np.random.randint(1, 3),
        )
    else:
        raise ValueError('Invalid level {}'.format(level))



"""Polygon modifier classes.

These classes are used to modify polygons (sprite vertex arrays, colors, and
opacities), typically before rendering. They can be used to adjust polygon
positions to render first-person, duplicate sprites when simulating torus
geometry, etc.
"""

import abc
import numpy as np


class AbstractPolygonModifier(abc.ABC):
    """Abstract polygon modifier class.
    
    All polygon modifiers must inherit from this class.
    """
    
    @abc.abstractmethod
    def __call__(self, state):
        """Get polygon modifier.
        
        Args:
            state: Environment state.

        Returns:
            Function taking in list of polygons and returning list of polygons.
        """
        pass


class DoNothing(AbstractPolygonModifier):
    """Does not modify polygons."""
    
    def __call__(self, state):
        def _do_nothing(_, sprite):
            return [(sprite.vertices, sprite.color, sprite.opacity)]
        return _do_nothing


class FirstPersonAgent(AbstractPolygonModifier):
    """Modifies polygon positions to put an agent sprite at [0.5, 0.5]."""

    def __init__(self, agent_layer):
        """Constructor.

        Args:
            first_person_agent: String. Will apply a translation such that the
                agent sprite (first sprite in this layer) is at position
                [0.5, 0.5].
        """
        self._agent_layer = agent_layer

    def __call__(self, state):
        """Get polygon modifier that translates lists of polygons."""
        agents = state[self._agent_layer]
        agent_position = agents[0].position
        delta_position = np.array([0.5, 0.5]) - agent_position
        def _sprite_to_polygons(layer, sprite):
            outs = [
                (sprite.vertices + delta_position, sprite.color, sprite.opacity)
            ]
            return outs
        return _sprite_to_polygons


class TorusGeometry(AbstractPolygonModifier):
    """Duplicates polygons at a grid of positions.
    
    Specifically, duplicates polygons at a 3x3 grid of positions centered around
    [0, 0] + polygon_position. This is used in torus environments to make a
    sprite appear to smoothly and incrementally disappear off one edge of the
    arena and reappear on the opposite edge.
    """

    def __init__(self, wrap_layers):
        """Constructor.

        Args:
            wrap_layers: String or iterable of strings. All sprites in these
                layers will be rendered as if the arena is a torus.
        """
        if not isinstance(wrap_layers, (list, tuple)):
            wrap_layers = [wrap_layers]
        self._wrap_layers = wrap_layers

    def __call__(self, state):
        """Get polygon modifier rendering sprites as if the arena is a torus."""
        del state
        def _sprite_to_polygons(layer, sprite):
            squared_polygons = [
                (sprite.vertices + np.array([i, j]), sprite.color,
                 sprite.opacity)
                for i in [-1., 0., 1.]
                for j in [-1., 0., 1.]
            ]
            return squared_polygons
        return _sprite_to_polygons



"""Contact prediction task.

In this task two red balls fall into a box. They bounce elastically off the
walls of the box, ultimately disappearing off the top of the screen because
there is no gravity. The subject's goal is to predict whether they will contact
each other. There is an occluder covering the bottom portion of the box. The
occluder may be translucent, depending on the argument to get_config(_).

The main entry point is the get_config() function at the bottom of this file.
"""

import collections
import numpy as np

from moog import action_spaces
from moog import game_rules
from moog import observers
from moog import physics as physics_lib
from moog import sprite
from moog import tasks
from moog.state_initialization import distributions as distribs


def _get_config(translucent_occluder):
    """Get environment config."""

    ############################################################################
    # Physics
    ############################################################################

    elastic_collision = physics_lib.Collision(
        elasticity=1., symmetric=False, update_angle_vel=False)
    physics = physics_lib.Physics(
        (elastic_collision, 'targets', 'walls'),
        updates_per_env_step=10,
    )

    def _predict_contact(state):
        """Predict whether targets will contact."""
        while True:
            if state['targets'][0].overlaps_sprite(state['targets'][1]):
                return True
            if all(s.y > 1.1 and s.y_vel > 0 for s in state['targets']):
                # Both targets above screen and moving up
                break
            physics.step(state)
        return False

    ############################################################################
    # Sprite initialization
    ############################################################################

    # Targets
    target_y_speed = 0.02
    target_factors = distribs.Product(
        [distribs.Continuous('x', 0.15, 0.85),
         distribs.Continuous('x_vel', -target_y_speed, target_y_speed)],
        y_vel=-target_y_speed, scale=0.16, shape='circle', opacity=192, c0=255,
        c1=0, c2=0,
    )

    # Occluder
    occluder = sprite.Sprite(
        x=0.5, y=0.2, shape='square', scale=1., c0=192, c1=192, c2=128,
        opacity=128 if translucent_occluder else 255
    )

    # Walls
    bottom_wall = [[-1, 0.1], [2, 0.1], [2, -1], [-1, -1]]
    left_wall = [[0.05, -1], [0.05, 4], [-1, 4], [-1, -1]]
    right_wall = [[0.95, -1], [0.95, 4], [2, 4], [2, -1]]
    walls = [
        sprite.Sprite(shape=np.array(v), x=0, y=0, c0=128, c1=128, c2=128)
        for v in [bottom_wall, left_wall, right_wall]
    ]
    
    # Make response boxes and tokens
    response_box_factors = dict(
        y=0.05, scale=0.12, shape='square', aspect_ratio=0.5, c0=0, c1=0, c2=0)
    response_boxes = [
        sprite.Sprite(x=0.4, **response_box_factors),
        sprite.Sprite(x=0.6, **response_box_factors),
    ]
    response_token_factors = dict(
        y=0.05, scale=0.03, shape='circle', c0=255, c1=0, c2=0, opacity=192)
    response_tokens = [
        sprite.Sprite(x=x, **response_token_factors)
        for x in [0.37, 0.43, 0.59, 0.61]
    ]
    
    def state_initializer():
        """Callable returning state ordereddict each episode reset."""
        agent = sprite.Sprite(
            x=0.5, y=0.05, scale=0.03, shape='spoke_4', c0=255, c1=255, c2=255)
        target_0 = sprite.Sprite(y=1.4, **target_factors.sample())
        target_1 = sprite.Sprite(
            y=np.random.uniform(1.7, 2.4), **target_factors.sample())
        screen = sprite.Sprite(
            x=0.5, y=0.5, shape='square', c0=128, c1=128, c2=128)

        state = collections.OrderedDict([
            ('targets', [target_0, target_1]),
            ('occluders', [occluder]),
            ('walls', walls),
            ('response_boxes', response_boxes),
            ('response_tokens', response_tokens),
            ('agent', [agent]),
            ('screen', [screen]),
        ])

        # Predict whether targets will contact, putting this information in
        # agent metadata
        orig_pos = [np.copy(s.position) for s in state['targets']]
        orig_vel = [np.copy(s.velocity) for s in state['targets']]
        agent.metadata = {'will_contact': _predict_contact(state)}
        for s, pos, vel in zip(state['targets'], orig_pos, orig_vel):
            s.position = pos
            s.velocity = vel

        return state

    ############################################################################
    # Task
    ############################################################################

    def _reward_fn(state):
        agent = state['agent'][0]
        if agent.overlaps_sprite(state['response_boxes'][0]):
            # Collision response
            return -1 if agent.metadata['will_contact'] else 1
        elif agent.overlaps_sprite(state['response_boxes'][1]):
            # No collision response
            return 1 if agent.metadata['will_contact'] else -1
        else:
            return 0
            
    conditional_task = tasks.Reset(
        condition=lambda state: _reward_fn(state) != 0,
        reward_fn=_reward_fn,
        steps_after_condition=5,
    )
    task = tasks.CompositeTask(conditional_task, timeout_steps=1000)

    ############################################################################
    # Action space
    ############################################################################

    action_space = action_spaces.Grid(
        scaling_factor=0.015,
        action_layers='agent',
        control_velocity=True,
    )

    ############################################################################
    # Observer
    ############################################################################

    observer = observers.PILRenderer(image_size=(64, 64), anti_aliasing=1)

    ############################################################################
    # Game rules
    ############################################################################

    screen_vanish = game_rules.VanishByFilter('screen')
    screen_vanish = game_rules.TimedRule(
        step_interval=(15, 16), rules=(screen_vanish,))

    rules = (screen_vanish,)

    ############################################################################
    # Final config
    ############################################################################

    config = {
        'state_initializer': state_initializer,
        'physics': physics,
        'task': task,
        'action_space': action_space,
        'observers': {'image': observer},
        'game_rules': rules,
    }
    return config


def get_config(translucent_occluder):
    return _get_config(translucent_occluder)



"""Task for resetting environment upon a condition of the environment state."""

from . import abstract_task
import inspect
import numpy as np


class Reset(abstract_task.AbstractTask):
    """Reset task.
    
    This task resets the environment when (or some fixed number of steps after)
    the environment state meets a condition. It can be used for example to reset
    once all prey are gone.

    There is also an option to specify a reward (as a function of the state) to
    be computed when the condition is met.
    """

    def __init__(self, condition, reward_fn=None, steps_after_condition=np.inf):
        """Constructor.

        Args:
            condition: Function with one of the following signatures:
                    * state --> bool
                    * state, meta_state --> bool
                The bool is whether to reset.
            reward_fn: Reward function taking in state and returning scalar
                reward. Only called if condition(state) is True.
            steps_after_condition: Int. Number of steps after condition is True
                to reset.
        """
        if len(inspect.signature(condition).parameters.values()) == 1:
            self._condition = lambda state, meta_state: condition(state)
        else:
            self._condition = condition
        self._steps_after_condition = steps_after_condition

        if reward_fn is None:
            reward_fn = lambda _: 0.
        self._reward_fn = reward_fn

    def reset(self, state, meta_state):
        # We reset to infinity, because self._steps_until_reset will be
        # decremented every time self.reward() is called, so is only set to a
        # finite value when the condition is met and reset is imminent.
        self._steps_until_reset = np.inf

    def reward(self, state, meta_state, step_count):
        """Compute reward."""
        del step_count
        if (self._steps_until_reset == np.inf and
                self._condition(state, meta_state)):
            reward = self._reward_fn(state)
            self._steps_until_reset = self._steps_after_condition
        else:
            reward = 0.

        self._steps_until_reset -= 1
        should_reset = self._steps_until_reset < 0

        return reward, should_reset



"""Environment.

This contains the Environment class. Evey MOOG task is an instance of this
Environment. It inherits from the dm_env.Environment API, but see
env_wrappers.gym_wrapper for an OpenAI Gym interface.
"""

import dm_env


class Environment(dm_env.Environment):
    """Environment class.

    Every object-oriented game is an instance of this class. This class pulls
    together all of the components of the game, such as the state initializer,
    physics (physics and forces), task (reward function), action space,
    observer (renderer), and game rules (game_rules). This is the only place
    where all of those components interact.

    The state of the environment is an OrderedDict containing all of physical
    objects (sprites) in the environment. The keys of the state are strings, and
    the values are iterables of sprite.Sprite instances. This state can be
    thought of as an ordered set of layers in the environment. This state is
    passed into physics and action space for updating, into the task for
    computing reward, and into the observer for rendering.
    """

    def __init__(self,
                 state_initializer,
                 physics,
                 task,
                 action_space,
                 observers,
                 game_rules=(),
                 meta_state_initializer=None):
        """Constructor.

        Args:
            state_initializer: Callable returning an initial state, which is an
                OrderedDict of iterables of sprite.Sprite instances. This is
                called at the beginning of each episode.
            physics: Instance of physics.AbstractPhysics. Must have methods:
                * reset(state)
                * step(state) --- in-place update the state each timestep.
            task: Instance of tasks.AbstractTask. Must have methods:
                * reset(state, meta_state)
                * reward(state, meta_state, step_count) returning scalar reward
                    and bool should_reset.
            action_space: Instance of action_spaces.AbstractActionSpace. Must
                have methods:
                * step(state, action)
                * reset(state)
            observers: Dict. Each value must be an instance of
                observers.AbstractObserver, hence callable taking in state and
                returning observation. The keys are the keys of an observation.
                For example, if you would like the environments' observations
                (returned in the timestep of each step) to contain multiple
                kinds of observations (e.g. a rendered image and a state
                description), you can let this observers argument be a
                dictionary {key_0: observer_0, key_1: observer_1} and the
                observation will be a dictionary {key_0: obs_0, key_1: obs_1}.
            game_rules: Iterable of instances of
                game_rules.AbstractRule. Each element is called on the state
                and meta_state every environment step and can modify them
                in-place.
            meta_state_initializer: Optional callable returning environment
                meta_state. If provided, is called every episode reset.
                Environment meta_state is only used by task and game rules.
        """
        self.state_initializer = state_initializer
        self.physics = physics
        self.task = task
        self.action_space = action_space
        self.observers = observers
        self.game_rules = game_rules

        if meta_state_initializer is None:
            self._meta_state_initializer = lambda: None
        else:
            self._meta_state_initializer = meta_state_initializer

    def reset(self):
        """Reset (start a new episode)."""
        self._reset_next_step = False
        self.step_count = 0
        
        self._state = self.state_initializer()
        self._meta_state = self._meta_state_initializer()
        self.task.reset(self._state, self._meta_state)
        self.physics.reset(self._state)
        self.action_space.reset(self._state)
        for rule in self.game_rules:
            rule.reset(self._state, self._meta_state)
            rule.step(self._state, self._meta_state)
        
        return dm_env.restart(self.observation())

    def step(self, action):
        """Step the environment with an action."""
        if self._reset_next_step:
            return self.reset()

        # Apply the game_rules
        for rule in self.game_rules:
            rule.step(self._state, self._meta_state)

        # Apply the action
        self.action_space.step(self._state, action)

        # Step the physics
        self.physics.step(self._state)

        # Compute reward
        self.step_count += 1
        reward, should_reset = self.task.reward(
            self._state, self._meta_state, self.step_count)

        # Take observation
        observation = self.observation()

        # Return transition
        if should_reset:
            self._reset_next_step = True
            return dm_env.termination(reward=reward, observation=observation)
        else:
            return dm_env.transition(reward=reward, observation=observation)

    def observation(self):
        """Returns a dictionary of observations."""
        return {k: observer(self._state)
                for k, observer in self.observers.items()}

    def observation_spec(self):
        """Returns a dictionary of dm_env specs for the observations."""
        observation_specs = {
            name: observer.observation_spec()
            for name, observer in self.observers.items()
        }
        return observation_specs

    def action_spec(self):
        """Returns the action space's .action_spec()."""
        return self.action_space.action_spec()

    @property
    def state(self):
        """State of environment."""
        return self._state
    
    @property
    def meta_state(self):
        """Meta-state of environment."""
        return self._meta_state

    @property
    def reset_next_step(self):
        """Whether to reset (start a new episode) on the next step."""
        return self._reset_next_step



""".. include:: README.md"""

from .abstract_rule import AbstractRule
from .change_layer import ChangeLayer
from .conditional import ConditionalRule
from .contact_rules import get_contact_counter
from .contact_rules import get_contact_indices
from .contact_rules import ModifyOnContact
from .create_sprites import CreateSprites
from .fixation import Fixation
from .modify_meta_state import ModifyMetaState
from .modify_meta_state import UpdateMetaStateValue
from .modify_sprites import ModifySprites
from .portal import Portal
from .re_center import KeepNearCenter
from .task_phases import Phase
from .task_phases import PhaseSequence
from .timing import DelayedRule
from .timing import TemporaryRule
from .timing import TimedRule
from .vanish import Vanish
from .vanish import VanishByFilter
from .vanish import VanishOnContact



"""Composite task."""

from . import abstract_task
import numpy as np


class CompositeTask(abstract_task.AbstractTask):
    """CompositeTask task.

    This combines multiple tasks at once, summing the rewards from each of them.
    This can be useful for example to have a predator/prey task where there are
    positive rewards for catching the prey and negative rewards for being caught
    by the predators.
    """

    def __init__(self, *tasks, timeout_steps=np.inf):
        """Constructor.

        Args:
            tasks: Tasks to compose. Reward will be the sum of the rewards from
                each of these tasks.
            timeout_steps: After this number of steps since reset, a reset is
                forced.
        """
        self._tasks = tasks
        self._timeout_steps = timeout_steps

    def reset(self, state, meta_state):
        for task in self._tasks:
            task.reset(state, meta_state)

    def reward(self, state, meta_state, step_count):
        """Compute reward."""
        reward = 0
        should_reset = step_count >= self._timeout_steps
        for task in self._tasks:
            task_reward, task_should_reset = task.reward(
                state, meta_state, step_count)
            reward += task_reward
            should_reset = should_reset or task_should_reset
        
        return reward, should_reset



"""Rules that control the timing of when to apply other rules.

These are useful for when you want to delay a rule of apply a rule only in a
specific timeframe.

If you want your task to have trial structure, consider looking at the rules in
task_phases.py. Those are more amenable to creating tasks with phases, e.g.
fixation phase -> stimulus phase -> response phase -> reward phase.
"""

from . import abstract_rule
import numpy as np


class TimedRule(abstract_rule.AbstractRule):
    """Apply a set of rules only during a specified time interval.
    
    The time interval step_interval is given in terms of number of environment
    steps, i.e. number of times this rule is called. The upper bound (second
    element) of step_interval may be infinite. The step_interval may also be a
    callable returning random intervals.

    For example, if you want some rule my_rule to step only between 20 and 40
    environment steps (i.e. 20 to 40 timesteps after each reset), you can use
    TimedRule((20, 40), my_rule).
    """

    def __init__(self, step_interval, rules):
        """Constructor.

        Args:
            step_interval: Length-2 iterable [start_step_count, stop_step_count]
                or callable returning such. This range specifies the interval in
                which to step rules. The units of measurement are times this
                rule is called --- this usually corresponds to environment
                steps, but may be offset if this TimedRule itself is the
                argument of some other TimedRule.
            rules: Rule or iterable of rules to step during the step_interval.
        """
        if not callable(step_interval):
            self._step_interval = lambda: step_interval
        else:
            self._step_interval = step_interval
        
        if not isinstance(rules, (list, tuple)):
            rules = (rules,)
        self._rules = rules

    def reset(self, state, meta_state):
        self._steps_until_start, self._steps_until_stop = self._step_interval()
        for rule in self._rules:
            rule.reset(state, meta_state)

    def step(self, state, meta_state):
        """Apply rule to state."""
        if self._steps_until_start <= 0 and self._steps_until_stop > 0:
            for rule in self._rules:
                rule.step(state, meta_state)
        self._steps_until_start -= 1
        self._steps_until_stop -= 1


class DelayedRule(TimedRule):
    """TimedRule that starts at specified time and never stops."""

    def __init__(self, steps_until_start, rules, duration=np.inf):
        """Constructor.

        Args:
            steps_until_start: Int or callable returning Int. This is the time
                at which to start stepping rules.
            rules: Rule or iterable of rules.
            duration: Optional int or callable. Duration of the rule.
        """
        if not callable(steps_until_start):
            callable_steps_until_start = lambda: steps_until_start
        else:
            callable_steps_until_start = steps_until_start

        if not callable(duration):
            callable_duration = lambda: duration
        else:
            callable_duration = duration

        def _step_interval():
            start_time = callable_steps_until_start()
            return (start_time, start_time + callable_duration())
        
        super().__init__(_step_interval, rules)


class TemporaryRule(TimedRule):
    """TimedRule that starts immediately but stops after specified duration."""

    def __init__(self, steps_until_stop, rules):
        """Constructor.

        Args:
            steps_until_stop: Int or callable returning Int. This is the time at
                which to stop stepping rules.
            rules: Rule or iterable of rules.
        """
        if not callable(steps_until_stop):
            step_interval = lambda: (0, steps_until_stop)
        else:
            step_interval = lambda: (0, steps_until_stop())
        super().__init__(step_interval, rules)



"""Environment wrapper class for running demo on multi-agent tasks.

The entry point is MultiAgentEnvironment, a class that behaves like a moog
environment but has agents living in it that take actions for all the agents
that the user is not controlling in the gui. So this effectively turns the
multi-agent environment into a single-agent demo-able environment.
"""

from moog import env_wrappers


class MultiAgentEnvironment(env_wrappers.AbstractEnvironmentWrapper):
    """Environment class for converting multi-agent into single-agent task."""
    
    def __init__(self, environment, agent_name, **other_agents):
        """Constructor.

        Args:
            environment: Instance of ../moog/environment.Environment. Should
                have a Composite action space.
            agent_name: Name of the agent being controlled by the player's
                joystick in the demo. Specifically, this is the key used for the
                joystick action value when passed into the Composite action
                space in the underlying environment.
            other_agents: Dictionary of agents. Keys must contain all the keys
                expected by the Composite action space except the agent_name
                key. Values are agent instances with a step(observation) method
                that returns an action.
        """
        super(MultiAgentEnvironment, self).__init__(environment)
        self._agent_name = agent_name
        self._other_agents = other_agents

    def step(self, action):
        """Step the environment with an action.
        
        This action should come from the demo. The actions for all other agents
        are filled in by stepping self._other_agents.
        """
        obs = self.observation()
        action_dict = {
            k: agent.step(obs) for k, agent in self._other_agents.items()
        }
        action_dict[self._agent_name] = action

        return self._environment.step(action_dict)



"""Random force."""

from . import abstract_force
import numpy as np


class RandomForce(abstract_force.AbstractNewtonianForce):
    """Random force class."""

    def __init__(self, max_force_magnitude):
        """Constructor.

        This class produces a force with magnitude uniformly sampled in
        [0, max_force_magnitude] and angle of direction randomly uniformly
        sampled in [0, 2 * pi].

        Args:
            max_force_magnitude: Scalar. Maximum force magnitude.
        """
        self._max_force_magnitude = max_force_magnitude

    def _compute_forces(self, sprite):
        r = np.random.uniform(0, self._max_force_magnitude)
        theta = np.random.uniform(0, 2 * np.pi)
        force = np.array([r * np.cos(theta), r * np.sin(theta)])
        return (force,)



"""Rules to facilitate tasks with phases.

This rules are useful if you want trials to have phasic structure, e.g.
fixation phase -> stimulus phase -> response phase -> reward phase,
with each transition between phases controlled by either some condition or a
temporal duration.
"""

from . import abstract_rule
import inspect
import numpy as np


class Phase():
    """Phase rule.
    
    This rule applies one-time-rules the first step, then continual rules for
    all subsequent steps until either an end condition is met or a timeout
    duration is reached.

    This is usually used as part of PhaseSequence() (see below).
    """

    def __init__(self,
                 one_time_rules=(),
                 continual_rules=(),
                 end_condition=None,
                 duration=np.inf,
                 name=''):
        """Constructor.

        Args:
            one_time_rules: Rule or iterable of rules, to be applied once the
                first time this phase is stepped.
            continual_rules: Rule or iterable of rules, to be applied each step
                during this phase.
            end_condition: Function with one of the following signatures:
                    * state --> bool
                    * state, meta_state --> bool
                Output bool is whether phase should end. Default is always
                False.
            duration: Int. maximum duration of phase.
            name: String. Optional name for the phase. This is sometimes used by
                PhaseSequence().
        """
        if not isinstance(one_time_rules, (list, tuple)):
            self._one_time_rules = (one_time_rules,)
        else:
            self._one_time_rules = one_time_rules

        if not isinstance(continual_rules, (list, tuple)):
            self._continual_rules = (continual_rules,)
        else:
            self._continual_rules = continual_rules
        
        if end_condition is None:
            self._end_condition = lambda state, meta_state: False
        elif len(inspect.signature(end_condition).parameters.values()) == 1:
            self._end_condition = lambda state, meta_state: end_condition(state)
        else:
            self._end_condition =  end_condition

        if not callable(duration):
            self._duration =  lambda: duration
        else:
            self._duration = duration

        self._name = name

    def reset(self, state, meta_state):
        """Reset at beginning of episode."""
        for rule in list(self._one_time_rules) + list(self._continual_rules):
            rule.reset(state=state, meta_state=meta_state)
        self._should_end = False
        self._step_count = 0
        self._current_duration = self._duration()

    def step(self, state, meta_state):
        """Step rule on environment state and meta_state."""
        if self.should_end:
            return
        
        if self._step_count == 0:
            for rule in self._one_time_rules:
                rule.step(state=state, meta_state=meta_state)
        
        for rule in self._continual_rules:
            rule.step(state=state, meta_state=meta_state)
        
        self._step_count += 1

        if (self._step_count >= self._current_duration or
                self._end_condition(state, meta_state)):
            self._should_end = True

    @property
    def should_end(self):
        return self._should_end
    
    @property
    def name(self):
        return self._name


class PhaseSequence():
    """PhaseSequence rule.
    
    This rule applies multiple Phase rules in sequence, applying each after the
    previous one ends.
    """

    def __init__(self, *single_phases, meta_state_phase_name_key=None):
        """Constructor.

        Args:
            *single_phases: Instances of Phase() (see above).
            meta_state_phase_name_key: Optional string. If given, the
                environment meta_state will contain the name of the current
                phase in this key, assuming it is a dictionary.
        """
        self._phases = single_phases
        self._meta_state_key = meta_state_phase_name_key

    def reset(self, state, meta_state):
        for phase in self._phases:
            phase.reset(state=state, meta_state=meta_state)
        self._current_phase_ind = 0
        self._current_phase = self._phases[0]
        if self._meta_state_key is not None:
            meta_state[self._meta_state_key] = self._current_phase.name

    def step(self, state, meta_state):
        if self._current_phase_ind >= len(self._phases):
            pass
        
        self._current_phase.step(state=state, meta_state=meta_state)
        if self._current_phase.should_end:
            self._current_phase_ind += 1
            self._current_phase = self._phases[self._current_phase_ind]
            if self._meta_state_key is not None:
                meta_state[self._meta_state_key] = self._current_phase.name



"""Rules and functions involving sprite contacts.

This file contains the ModifyOnContact game rule, which modifies sprites if they
come in contact.

It also contains the get_contact_counter() function, which can be useful for
counting how many contacts there are between layers, e.g. for use in a
ConditionalRule.
"""

from . import abstract_rule
import itertools


def get_contact_indices(layer_0, layer_1):
    """Get counter that finds index pairs of all contacts between layers.
    
    Args:
        layer_0: String. Must be key in state.
        layer_1: String. Must be key in state.

    Returns:
        _call: Function state --> list, where the elements of the returned list
            are tuples (i_0, i_0) of indices of sprites in layer_0 and layer_1
            that are contacting.
    """

    def _call(state):
        """Gets all (i_0, i_1) such that layer_0[i_0] contacts layer_1[i_1]."""
        contact_indices = []
        for i_0, sprite_0 in enumerate(state[layer_0]):
            for i_1, sprite_1 in enumerate(state[layer_1]):
                if sprite_0.overlaps_sprite(sprite_1):
                    contact_indices.append((i_0, i_1))
        return contact_indices
    return _call


def get_contact_counter(layer_0, layer_1):
    """Get counter that finds how many contacts there are between layers.
    
    Args:
        layer_0: String. Must be key in state.
        layer_1: String. Must be key in state.

    Returns:
        Function state --> int returning how many contacts there are between
            layer_0 and layer_1.
    """
    _get_contact_indices = get_contact_indices(layer_0, layer_1)
    return lambda state: len(_get_contact_indices(state))


class ModifyOnContact(abstract_rule.AbstractRule):
    """Modify sprites if they contact each other."""

    def __init__(self,
                 layers_0,
                 layers_1,
                 modifier_0=None,
                 modifier_1=None,
                 filter_0=None,
                 filter_1=None,):
        """Constructor.

        Applies modifier_0 to those sprites in layers_0 that satisfy filter_0
        and contact a sprite in layers_1. Similarly, applies modifier_1 to those
        sprites in layers_1 that satisfy filter_1 and contact a sprite in
        layers_0.
        
        Args:
            layers_0: String or iterable of strings. Must be layer name(s) in
                environment state.
            layers_1: String or iterable of strings. Must be layer name(s) in
                environment state.
            modifier_0: Function taking in a sprite and modifying in place.
            modifier_1: Function taking in a sprite and modifying in place.
            filter_0: Function taking in a sprite and returning bool.
            filter_1: Function taking in a sprite and returning bool.
        """
        if not isinstance(layers_0, (list, tuple)):
            layers_0 = (layers_0,)
        self._layers_0 = layers_0
        if not isinstance(layers_1, (list, tuple)):
            layers_1 = (layers_1,)
        self._layers_1 = layers_1
        
        self._modifier_0 = modifier_0
        self._modifier_1 = modifier_1

        if filter_0 is None:
            filter_0 = lambda x: True
        self._filter_0 = filter_0
        if filter_1 is None:
            filter_1 = lambda x: True
        self._filter_1 = filter_1

    def _modify_asymmetric(self, sprites_modifying, sprites_contacting,
                           modifier, filter):
        """Modify sprites_modifying if contcating sprites_contacting."""
        if modifier is not None:
            for s in sprites_modifying:
                if not filter(s):
                    continue
                contacts = [s.overlaps_sprite(x) for x in sprites_contacting
                            if id(x) != id(s)]
                if any(contacts):
                    modifier(s)

    def step(self, state, meta_state):
        """Apply rule to state."""
        del meta_state
        
        sprites_0 = list(itertools.chain(*[state[k] for k in self._layers_0]))
        sprites_1 = list(itertools.chain(*[state[k] for k in self._layers_1]))
        
        self._modify_asymmetric(
            sprites_0, sprites_1, self._modifier_0, self._filter_0)
        self._modify_asymmetric(
            sprites_1, sprites_0, self._modifier_1, self._filter_1)



"""Environment wrapper class that supports mental simulation.

This wrapper is useful when running model-based RL algorithms, because it
provides a ground-truth model of the environment.

Note: If there is stochasticity in the environment (e.g. maybe your physics or
game rules have randomness), this will only take samples at each simulation
step, not representing the full distribution of possible simulations. You can
step and pop many times to estimate the distribution of possibilities.
"""

import copy
from moog import env_wrappers


class SimulationEnvironment(env_wrappers.AbstractEnvironmentWrapper):
    """Environment class supporting mental simulation.
    
    This wrapper supports the normal environment interface, but adds two
    additional methods:
        * sim_step(action): This takes a step in mental simulation, advancing
            the state and returning a timestep. However, it also keeps a
            representation of the current state in a stack, to allow for
            restoring later.
        * sim_pop(index=-1): This restores to a previous state in the mental
            simulation stack. Specifically, it restores to the state at level
            index. For example, if we've called sim_step() two times, then
            sim_pop(-1) or sim_pop(1) would restore to the state after the first
            sim_step() call, whereas sim_pop(0) would restore the original state
            before both sim_step() calls.
    
    You can think of sim_step() and sim_pop() as allowing traversals through a
    tree of mental simulation.
    """
    
    def __init__(self, environment):
        """Constructor.

        Args:
            environment: Instance of ../moog/environment.Environment.
        """
        super(SimulationEnvironment, self).__init__(environment)

    def reset(self):
        self.stack = []
        return super(SimulationEnvironment, self).reset()

    def step(self, action):
        """Step the environment with an action."""
        if self.stack:
            self.sim_pop(index=0)
        self.stack = []
        return super(SimulationEnvironment, self).step(action)
    
    def sim_step(self, action):
        """Take a simulation step of the environment with an action."""
        if self._environment.reset_next_step:
            # Should not simulate across episode boundaries
            return None
        self.stack.append({
            'state': copy.deepcopy(self._environment.state),
            'meta_state': copy.deepcopy(self._environment.meta_state),
            'action_space': copy.deepcopy(self._environment.action_space),
            'physics': copy.deepcopy(self._environment.physics),
            'game_rules': copy.deepcopy(self._environment.game_rules),
            'task': copy.deepcopy(self._environment.task),
            'step_count': copy.deepcopy(self._environment.step_count),
            'reset_next_step': copy.deepcopy(self._environment.reset_next_step),
        })
        return super(SimulationEnvironment, self).step(action)
    
    def sim_pop(self, index=-1):
        """Pop and restore the state at index off the stack."""
        restore_data = self.stack[index]
        self._environment._state = restore_data['state']
        self._environment.action_space = restore_data['action_space']
        self._environment.physics = restore_data['physics']
        self._environment.game_rules = restore_data['game_rules']
        self._environment.task = restore_data['task']
        self._environment._meta_state = restore_data['meta_state']  #pylint: disable=protected-access
        self._environment.step_count = restore_data['step_count']
        self._environment._reset_next_step = restore_data['reset_next_step']  #pylint: disable=protected-access
        self.stack = self.stack[:index]



"""Task for receiving rewards upon contact."""

from . import abstract_task
import inspect
import numpy as np


class ContactReward(abstract_task.AbstractTask):
    """ContactReward task.
    
    In this task if any sprite in layers_0 contacts any sprite in layers_1, a
    reward is given. Otherwise the reward is zero. Optionally, the task resets
    upon such a contact.

    This can be used for any contact-based reward, such as prey-seeking and
    predator-avoidance.
    """

    def __init__(self,
                 reward_fn,
                 layers_0,
                 layers_1,
                 condition=None,
                 reset_steps_after_contact=np.inf):
        """Constructor.

        Args:
            reward_fn: Scalar or function (sprite_0, sprite_1) --> scalar. If
                function, sprite_0 and sprite_1 are sprites in layers_0 and
                layers_1 respectively.
            layers_0: String or iterable of strings. Reward is given if a sprite
                in this layer(s) contacts a sprite in layers_1.
            layers_1: String or iterable of strings. Reward is given if a sprite
                in this layer(s) contacts a sprite in layers_0.
            condition: Optional condition function. If specified, must have one
                of the following signatures:
                    * sprite_0, sprite_1 --> bool
                    * sprite_0, sprite_1, meta_state --> bool
                The bool is whether to apply reward for those sprites
                contacting.
            reset_steps_after_contact: Int. How many steps after a contact to
                reset the environment. Defaults to infinity, i.e. never
                resetting.
        """
        if not callable(reward_fn):
            self._reward_fn = lambda sprite_0, sprite_1: reward_fn
        else:
            self._reward_fn = reward_fn
        
        if not isinstance(layers_0, (list, tuple)):
            layers_0 = [layers_0]
        self._layers_0 = layers_0

        if not isinstance(layers_1, (list, tuple)):
            layers_1 = [layers_1]
        self._layers_1 = layers_1

        if condition is None:
            self._condition = lambda s_agent, s_target, meta_state: True
        elif len(inspect.signature(condition).parameters.values()) == 2:
            self._condition = lambda s_a, s_t, meta_state: condition(s_a, s_t)
        else:
            self._condition = condition

        self._reset_steps_after_contact = reset_steps_after_contact

    def reset(self, state, meta_state):
        self._steps_until_reset = np.inf

    def reward(self, state, meta_state, step_count):
        """Compute reward.
        
        If any sprite_0 in self._layers_0 overlaps any sprite_1 in
        self._layers_1 and if self._condition(sprite_0, sprite_1, meta_state) is
        True, then the reward is self._reward_fn(sprite_0, sprite_1).

        Args:
            state: OrderedDict of sprites. Environment state.
            meta_state: Environment state. Unconstrained type.
            step_count: Int. Environment step count.

        Returns:
            reward: Scalar reward.
            should_reset: Bool. Whether to reset task.
        """
        reward = 0
        sprites_0 = [s for k in self._layers_0 for s in state[k]]
        sprites_1 = [s for k in self._layers_1 for s in state[k]]
        for s_0 in sprites_0:
            for s_1 in sprites_1:
                if not self._condition(s_0, s_1, meta_state):
                    continue
                if s_0.overlaps_sprite(s_1):
                    reward = self._reward_fn(s_0, s_1)
                    if self._steps_until_reset == np.inf:
                        self._steps_until_reset = (
                            self._reset_steps_after_contact)

        self._steps_until_reset -= 1
        should_reset = self._steps_until_reset < 0

        return reward, should_reset
